# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import sys

import pytest

from awscli.customizations.s3 import transferconfig


class TestTransferConfig:
    def build_config_with(self, **config_from_user):
        return transferconfig.RuntimeConfig().build_config(**config_from_user)

    def test_user_provides_no_config_uses_default(self):
        # If the user does not provide any config overrides,
        # we should just use the default values defined in
        # the module.
        config = transferconfig.RuntimeConfig()
        runtime_config = config.build_config()
        assert runtime_config == transferconfig.DEFAULTS

    def test_user_provides_partial_overrides(self):
        config_from_user = {
            'max_concurrent_requests': '20',
            'multipart_threshold': str(64 * (1024**2)),
        }
        runtime_config = self.build_config_with(**config_from_user)
        # Our overrides were accepted.
        assert runtime_config['multipart_threshold'] == int(
            config_from_user['multipart_threshold']
        )
        assert runtime_config['max_concurrent_requests'] == int(
            config_from_user['max_concurrent_requests']
        )
        # And defaults were used for values not specified.
        assert runtime_config['max_queue_size'] == int(
            transferconfig.DEFAULTS['max_queue_size']
        )

    def test_validates_integer_types(self):
        with pytest.raises(transferconfig.InvalidConfigError):
            self.build_config_with(max_concurrent_requests="not an int")

    def test_validates_positive_integers(self):
        with pytest.raises(transferconfig.InvalidConfigError):
            self.build_config_with(max_concurrent_requests="-10")

    def test_min_value(self):
        with pytest.raises(transferconfig.InvalidConfigError):
            self.build_config_with(max_concurrent_requests="0")

    def test_human_readable_sizes_converted_to_bytes(self):
        runtime_config = self.build_config_with(multipart_threshold="10MB")
        assert runtime_config['multipart_threshold'] == 10 * 1024 * 1024

    def test_long_value(self):
        # MAXSIZE is the max size of an int on python 2 and the maximum size
        # of Py_ssize_t on python 3, but notably not the maximum size of an
        # int since they are effectively unbounded.
        long_value = sys.maxsize + 1
        runtime_config = self.build_config_with(multipart_threshold=long_value)
        assert runtime_config['multipart_threshold'] == long_value

    @pytest.mark.parametrize(
        'provided,resolved',
        [
            (None, 'auto'),
            ('auto', 'auto'),
            ('classic', 'classic'),
            ('default', 'classic'),
            ('crt', 'crt'),
        ],
    )
    def test_set_preferred_transfer_client(self, provided, resolved):
        config_kwargs = {}
        if provided is not None:
            config_kwargs['preferred_transfer_client'] = provided
        runtime_config = self.build_config_with(**config_kwargs)
        assert runtime_config['preferred_transfer_client'] == resolved

    @pytest.mark.parametrize(
        'config_name,provided,expected',
        [
            # max_bandwidth cases
            ('max_bandwidth', '1MB/s', 1024 * 1024),
            ('max_bandwidth', '8Mb/s', 1024 * 1024),
            ('max_bandwidth', '1000', 1000),
            ('max_bandwidth', '1000B/s', 1000),
            ('max_bandwidth', '8000b/s', 1000),
            # target_bandwidth cases
            ('target_bandwidth', '5MB/s', 5 * 1024 * 1024),
            ('target_bandwidth', '1Mb/s', 1 * 1024 * 1024 / 8),
            ('target_bandwidth', '1000', 1000),
            ('target_bandwidth', '1000B/s', 1000),
            ('target_bandwidth', '8000b/s', 1000),
            # disk_throughput cases
            ('disk_throughput', '1MB/s', 1024 * 1024),
            ('disk_throughput', '10Mb/s', 10 * 1024 * 1024 / 8),
            ('disk_throughput', '1000', 1000),
            ('disk_throughput', '1000B/s', 1000),
            ('disk_throughput', '8000b/s', 1000),
        ],
    )
    def test_rate_conversions(self, config_name, provided, expected):
        params = {config_name: provided}
        runtime_config = self.build_config_with(**params)
        assert runtime_config[config_name] == expected

    @pytest.mark.parametrize(
        'config_name,provided',
        [
            # max_bandwidth cases
            ('max_bandwidth', '1MB'),
            ('max_bandwidth', '1B'),
            ('max_bandwidth', '1b'),
            ('max_bandwidth', '100/s'),
            ('max_bandwidth', ''),
            ('max_bandwidth', 'value-with-no-digits'),
            # target_bandwidth cases
            ('target_bandwidth', '1MB'),
            ('target_bandwidth', '1B'),
            ('target_bandwidth', '1b'),
            ('target_bandwidth', '100/s'),
            ('target_bandwidth', ''),
            ('target_bandwidth', 'value-with-no-digits'),
            # disk_throughput cases
            ('disk_throughput', '1MB'),
            ('disk_throughput', '1B'),
            ('disk_throughput', '1b'),
            ('disk_throughput', '100/s'),
            ('disk_throughput', ''),
            ('disk_throughput', 'value-with-no-digits'),
        ],
    )
    def test_invalid_rate_values(self, config_name, provided):
        params = {config_name: provided}
        with pytest.raises(transferconfig.InvalidConfigError):
            self.build_config_with(**params)

    def test_validates_preferred_transfer_client_choices(self):
        with pytest.raises(transferconfig.InvalidConfigError):
            self.build_config_with(preferred_transfer_client='not-supported')

    @pytest.mark.parametrize(
        'attr,val,expected',
        [
            ('should_stream', 'true', True),
            ('should_stream', 'false', False),
            ('should_stream', None, None),
            ('direct_io', 'true', True),
            ('direct_io', 'false', False),
            ('direct_io', None, None),
        ],
    )
    def test_convert_booleans(self, attr, val, expected):
        params = {attr: val}
        runtime_config = self.build_config_with(**params)
        assert runtime_config[attr] == expected


class TestConvertToS3TransferConfig:
    def test_convert(self):
        runtime_config = {
            'multipart_threshold': 1,
            'multipart_chunksize': 2,
            'max_concurrent_requests': 3,
            'max_queue_size': 4,
            'max_bandwidth': 1024 * 1024,
            'addressing_style': 'path',
            'use_accelerate_endpoint': True,
            'io_chunksize': 1024 * 1024,
            # This is a TransferConfig only option, it should
            # just be ignored if it's in the ~/.aws/config for now.
            'max_in_memory_upload_chunks': 1000,
        }
        result = transferconfig.create_transfer_config_from_runtime_config(
            runtime_config
        )
        assert result.multipart_threshold == 1
        assert result.multipart_chunksize == 2
        assert result.max_request_concurrency == 3
        assert result.max_request_queue_size == 4
        assert result.max_bandwidth == 1024 * 1024
        assert result.io_chunksize == 1024 * 1024
        assert result.max_in_memory_upload_chunks != 1000



# Copyright 2026 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.utils import alias_command


def register_rename_otel_commands(event_emitter):
    event_emitter.register(
        'building-command-table.cloudwatch',
        rename_otel_commands
    )


def rename_otel_commands(command_table, **kwargs):
    """Rename o-tel commands to otel, keeping o-tel as hidden aliases."""
    renames = {
        'get-o-tel-enrichment': 'get-otel-enrichment',
        'start-o-tel-enrichment': 'start-otel-enrichment',
        'stop-o-tel-enrichment': 'stop-otel-enrichment',
    }
    for old_name, new_name in renames.items():
        if old_name in command_table:
            alias_command(command_table, old_name, new_name)



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os


class FakeTransferFuture:
    def __init__(self, result=None, exception=None, meta=None):
        self._result = result
        self._exception = exception
        self.meta = meta

    def result(self):
        if self._exception:
            raise self._exception
        return self._result

    def set_exception(self, exception):
        self._exception = exception


class FakeTransferFutureMeta:
    def __init__(
        self, size=None, call_args=None, transfer_id=None, user_context=None
    ):
        self.size = size
        self.call_args = call_args
        self.transfer_id = transfer_id
        self.user_context = user_context


class FakeTransferFutureCallArgs:
    def __init__(self, **kwargs):
        for kwarg, val in kwargs.items():
            setattr(self, kwarg, val)


def make_loc_files(file_creator, size=None):
    """
    This sets up the test by making a directory named some_directory.  It
    has the file text1.txt and the directory another_directory inside.  Inside
    of another_directory it creates the file text2.txt.
    """
    if size:
        body = "*" * size
    else:
        body = 'This is a test.'

    filename1 = file_creator.create_file(
        os.path.join('some_directory', 'text1.txt'), body
    )

    filename2 = file_creator.create_file(
        os.path.join('some_directory', 'another_directory', 'text2.txt'), body
    )
    filename1 = str(filename1)
    filename2 = str(filename2)
    return [
        filename1,
        filename2,
        os.path.dirname(filename2),
        os.path.dirname(filename1),
    ]


def clean_loc_files(file_creator):
    """
    Removes all of the local files made.
    """
    file_creator.remove_all()


def compare_files(result_file, ref_file):
    """
    Ensures that the FileStat's properties are what they
    are suppose to be.
    """
    assert result_file.src == ref_file.src
    assert result_file.dest == ref_file.dest
    assert result_file.compare_key == ref_file.compare_key
    assert result_file.size == ref_file.size
    assert result_file.last_update == ref_file.last_update
    assert result_file.src_type == ref_file.src_type
    assert result_file.dest_type == ref_file.dest_type
    assert result_file.operation_name == ref_file.operation_name



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/

import asyncio
import time
from unittest.mock import Mock, patch

import pytest
from botocore.exceptions import ClientError

from awscli.customizations.ecs.expressgateway.display_strategy import (
    DisplayStrategy,
    InteractiveDisplayStrategy,
    TextOnlyDisplayStrategy,
)


class TestDisplayStrategy:
    """Test base DisplayStrategy class."""

    def test_base_strategy_not_implemented(self):
        """Test base class raises NotImplementedError."""
        strategy = DisplayStrategy()
        with pytest.raises(NotImplementedError):
            strategy.execute_monitoring(None, None, None)


@pytest.fixture
def mock_display():
    """Fixture that creates a mock display for testing."""

    async def mock_run_async():
        await asyncio.sleep(0.01)

    display = Mock()
    display.display = Mock()
    display.run = Mock(return_value=mock_run_async())
    return display


class TestInteractiveDisplayStrategy:
    """Test InteractiveDisplayStrategy."""

    @patch('time.sleep')
    def test_execute_with_mock_display(
        self, mock_sleep, ptk_app_session, mock_display
    ):
        """Test strategy executes with mocked display."""
        mock_collector = Mock()
        mock_collector.get_current_view = Mock(
            return_value="Test output {SPINNER}"
        )

        strategy = InteractiveDisplayStrategy(
            display=mock_display, use_color=True
        )

        mock_sleep.side_effect = KeyboardInterrupt()

        start_time = time.time()
        strategy.execute_monitoring(
            mock_collector, start_time, timeout_minutes=1
        )

        # Verify display was called
        assert mock_display.display.called
        assert mock_display.run.called

    def test_strategy_uses_provided_color_setting(self):
        """Test strategy respects use_color parameter."""
        mock_display = Mock()

        strategy_with_color = InteractiveDisplayStrategy(
            display=mock_display, use_color=True
        )
        assert strategy_with_color.use_color is True

        strategy_no_color = InteractiveDisplayStrategy(
            display=mock_display, use_color=False
        )
        assert strategy_no_color.use_color is False

    @patch('time.sleep')
    def test_completion_message_on_normal_exit(
        self, mock_sleep, ptk_app_session, mock_display, capsys
    ):
        """Test displays completion message when monitoring completes normally."""
        mock_collector = Mock()
        mock_collector.get_current_view = Mock(return_value="Resources ready")

        strategy = InteractiveDisplayStrategy(
            display=mock_display, use_color=True
        )

        mock_sleep.side_effect = KeyboardInterrupt()

        start_time = time.time()
        strategy.execute_monitoring(
            mock_collector, start_time, timeout_minutes=1
        )

        captured = capsys.readouterr()
        assert "Monitoring Complete!" in captured.out
        assert "Monitoring timed out!" not in captured.out

    @patch('time.sleep')
    def test_collector_output_is_displayed(
        self, mock_sleep, ptk_app_session, mock_display, capsys
    ):
        """Test that collector output appears in final output."""
        mock_collector = Mock()
        unique_output = "LoadBalancer lb-12345 ACTIVE"
        mock_collector.get_current_view = Mock(return_value=unique_output)

        strategy = InteractiveDisplayStrategy(
            display=mock_display, use_color=True
        )

        mock_sleep.side_effect = KeyboardInterrupt()

        start_time = time.time()
        strategy.execute_monitoring(
            mock_collector, start_time, timeout_minutes=1
        )

        captured = capsys.readouterr()
        assert unique_output in captured.out

    @patch('time.sleep')
    def test_execute_handles_service_inactive(
        self, mock_sleep, ptk_app_session, mock_display, capsys
    ):
        """Test strategy handles service inactive error."""
        mock_collector = Mock()
        error = ClientError(
            error_response={
                'Error': {
                    'Code': 'InvalidParameterException',
                    'Message': 'Cannot call DescribeServiceRevisions for a service that is INACTIVE',
                }
            },
            operation_name='DescribeServiceRevisions',
        )
        mock_collector.get_current_view = Mock(side_effect=error)

        strategy = InteractiveDisplayStrategy(
            display=mock_display, use_color=True
        )

        mock_sleep.side_effect = KeyboardInterrupt()

        start_time = time.time()
        strategy.execute_monitoring(
            mock_collector, start_time, timeout_minutes=1
        )

        # Strategy should handle the error and set output to "Service is inactive"
        captured = capsys.readouterr()
        assert "Service is inactive" in captured.out

    @patch('time.sleep')
    def test_execute_other_client_errors_propagate(
        self, mock_sleep, ptk_app_session, mock_display
    ):
        """Test strategy propagates non-service-inactive ClientErrors."""
        mock_collector = Mock()
        error = ClientError(
            error_response={
                'Error': {
                    'Code': 'AccessDeniedException',
                    'Message': 'Access denied',
                }
            },
            operation_name='DescribeServiceRevisions',
        )
        mock_collector.get_current_view = Mock(side_effect=error)

        strategy = InteractiveDisplayStrategy(
            display=mock_display, use_color=True
        )

        mock_sleep.side_effect = KeyboardInterrupt()

        start_time = time.time()

        # Other client errors should propagate
        with pytest.raises(ClientError) as exc_info:
            strategy.execute_monitoring(
                mock_collector, start_time, timeout_minutes=1
            )

        assert (
            exc_info.value.response['Error']['Code'] == 'AccessDeniedException'
        )

    @patch('time.sleep')
    def test_display_cleanup_on_exception(
        self, mock_sleep, ptk_app_session, mock_display
    ):
        """Test display app is properly shut down when exception occurs."""
        mock_collector = Mock()
        error = ClientError(
            error_response={'Error': {'Code': 'ThrottlingException'}},
            operation_name='DescribeServiceRevisions',
        )
        mock_collector.get_current_view = Mock(side_effect=error)

        strategy = InteractiveDisplayStrategy(
            display=mock_display, use_color=True
        )
        mock_sleep.side_effect = KeyboardInterrupt()

        with pytest.raises(ClientError):
            strategy.execute_monitoring(
                mock_collector, time.time(), timeout_minutes=1
            )

        # Verify app.exit() was called in finally block despite exception
        mock_display.app.exit.assert_called()


class TestTextOnlyDisplayStrategy:
    """Test TextOnlyDisplayStrategy."""

    @patch('time.sleep')
    def test_execute_with_mock_collector(self, mock_sleep, capsys):
        """Test strategy executes sync loop with text output."""
        mock_collector = Mock()
        mock_collector.get_current_view = Mock(return_value="Test output")
        mock_collector.cached_monitor_result = (None, "Test info")

        strategy = TextOnlyDisplayStrategy(use_color=True)

        # Make sleep raise to exit loop after first iteration
        mock_sleep.side_effect = KeyboardInterrupt()

        start_time = time.time()
        strategy.execute_monitoring(
            mock_collector, start_time, timeout_minutes=1
        )

        output = capsys.readouterr().out
        printed_output = output
        assert "Starting monitoring" in printed_output
        assert "stopped by user" in printed_output
        assert "complete" in printed_output

    @patch('time.sleep')
    @patch('time.time')
    def test_execute_handles_timeout(self, mock_time, mock_sleep, capsys):
        """Test strategy handles timeout correctly."""
        mock_collector = Mock()
        mock_collector.get_current_view = Mock(return_value="Test output")
        mock_collector.cached_monitor_result = (None, None)

        strategy = TextOnlyDisplayStrategy(use_color=True)

        # Simulate timeout after first poll
        start_time = 1000.0
        mock_time.side_effect = [
            1000.0,  # First check - within timeout
            2000.0,  # Second check - exceeded timeout
        ]

        strategy.execute_monitoring(
            mock_collector, start_time, timeout_minutes=1
        )

        output = capsys.readouterr().out
        printed_output = output
        assert "timeout reached" in printed_output.lower()

    def test_strategy_uses_provided_color_setting(self):
        """Test strategy respects use_color parameter."""
        strategy_with_color = TextOnlyDisplayStrategy(use_color=True)
        assert strategy_with_color.stream_display.use_color is True

        strategy_no_color = TextOnlyDisplayStrategy(use_color=False)
        assert strategy_no_color.stream_display.use_color is False



#!/usr/bin/env python
# Copyright 2012-2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import awscli.clidriver
from awscli.testutils import BaseAWSCommandParamsTest


class TestPurgeQueue(BaseAWSCommandParamsTest):
    prefix = 'sqs purge-queue'

    def test_simple(self):
        url = 'https://foo.bar/queue'
        cmdline = self.prefix
        cmdline += ' --queue-url %s' % url
        result = {'QueueUrl': url}
        self.assert_params_for_cmd(cmdline, result)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os

from awscli.testutils import BaseAWSCommandParamsTest

PASSWORD_DATA = (
    "GWDnuoj/7pbMQkg125E8oGMUVCI+r98sGbFFl8SX+dEYxMZzz+byYwwjvyg8i"
    "SGKaLuLTIWatWopVu5cMWDKH65U4YFL2g3LqyajBrCFnuSE1piTeS/rPQpoSv"
    "BN5FGj9HWqNrglWAJgh9OZNSGgpEojBenL/0rwSpDWL7f/f52M5doYA6q+v0y"
    "gEoi1Wq6hcmrBfyA4seW1RlKgnUru5Y9oc1hFHi53E3b1EkjGqCsCemVUwumB"
    "j8uwCLJRaMcqrCxK1smtAsiSqk0Jk9jpN2vcQgnMPypEdmEEXyWHwq55fjy6c"
    "h+sqYcwumIL5QcFW2JQ5+XBEoFhC66gOsAXow=="
)


class TestGetPasswordData(BaseAWSCommandParamsTest):
    prefix = 'ec2 get-password-data'

    def setUp(self):
        super(TestGetPasswordData, self).setUp()
        self.parsed_response = {
            'InstanceId': 'i-12345678',
            'Timestamp': '2013-07-27T18:29:23.000Z',
            'PasswordData': PASSWORD_DATA,
        }

    def test_no_priv_launch_key(self):
        args = ' --instance-id i-12345678'
        cmdline = self.prefix + args
        result = {'InstanceId': 'i-12345678'}
        output = self.assert_params_for_cmd(cmdline, result, expected_rc=0)[0]
        self.assertIn('"InstanceId": "i-12345678"', output)
        self.assertIn('"Timestamp": "2013-07-27T18:29:23.000Z"', output)
        self.assertIn('"PasswordData": "%s"' % PASSWORD_DATA, output)

    def test_nonexistent_priv_launch_key(self):
        args = ' --instance-id i-12345678 --priv-launch-key foo.pem'
        cmdline = self.prefix + args
        error_msg = self.assert_params_for_cmd(cmdline, expected_rc=255)[1]
        self.assertIn(
            'priv-launch-key should be a path to '
            'the local SSH private key file used '
            'to launch the instance.\n',
            error_msg,
        )

    def test_priv_launch_key(self):
        key_path = os.path.join(os.path.dirname(__file__), 'testcli.pem')
        args = ' --instance-id i-12345678 --priv-launch-key %s' % key_path
        cmdline = self.prefix + args
        result = {'InstanceId': 'i-12345678'}
        output = self.assert_params_for_cmd(cmdline, result, expected_rc=0)[0]
        self.assertIn('"InstanceId": "i-12345678"', output)
        self.assertIn('"Timestamp": "2013-07-27T18:29:23.000Z"', output)
        self.assertIn('"PasswordData": "=mG8.r$o-s"', output)



import contextlib
import glob
import json
import os
import platform
import shutil
import subprocess
import sys
import tempfile
import zipfile


class BadRCError(Exception):
    pass


def run(cmd, cwd=None, env=None, echo=True):
    if echo:
        sys.stdout.write("Running cmd: %s\n" % cmd)
    kwargs = {
        'shell': True,
        'stdout': subprocess.PIPE,
        'stderr': subprocess.PIPE,
    }
    if isinstance(cmd, list):
        kwargs['shell'] = False
    if cwd is not None:
        kwargs['cwd'] = cwd
    if env is not None:
        kwargs['env'] = env
    p = subprocess.Popen(cmd, **kwargs)
    stdout, stderr = p.communicate()
    output = stdout.decode('utf-8') + stderr.decode('utf-8')
    if p.returncode != 0:
        raise BadRCError(
            "Bad rc (%s) for cmd '%s': %s" % (p.returncode, cmd, output)
        )
    return output


def extract_zip(zipfile_name, target_dir):
    with zipfile.ZipFile(zipfile_name, 'r') as zf:
        for zf_info in zf.infolist():
            # Works around extractall not preserving file permissions:
            # https://bugs.python.org/issue15795
            extracted_path = zf.extract(zf_info, target_dir)
            os.chmod(extracted_path, zf_info.external_attr >> 16)


@contextlib.contextmanager
def tmp_dir():
    dirname = tempfile.mkdtemp()
    try:
        yield dirname
    finally:
        shutil.rmtree(dirname)


@contextlib.contextmanager
def cd(dirname):
    original = os.getcwd()
    os.chdir(dirname)
    try:
        yield
    finally:
        os.chdir(original)


def bin_path():
    """Get the system's binary path, either `bin` on reasonable systems
    or `Scripts` on Windows.
    """
    path = "bin"

    if platform.system() == "Windows":
        path = "Scripts"

    return path


def virtualenv_enabled():
    # Helper function to see if we need to make
    # our own virtualenv for installs.
    return bool(os.environ.get('VIRTUAL_ENV'))


def update_metadata(dirname, **kwargs):
    print('Update metadata values %s' % kwargs)
    metadata_file = os.path.join(dirname, 'awscli', 'data', 'metadata.json')
    with open(metadata_file) as f:
        metadata = json.load(f)
    for key, value in kwargs.items():
        metadata[key] = value
    with open(metadata_file, 'w') as f:
        json.dump(metadata, f)


def save_to_zip(dirname, zipfile_name):
    if zipfile_name.endswith('.zip'):
        zipfile_name = zipfile_name[:-4]
    shutil.make_archive(zipfile_name, 'zip', dirname)


def remove_dist_info_directories(dirname, keep_packages):
    """Remove .dist-info directories except for specified packages.
    Generally these are not needed in the final CLI executable,
    unless specified via keep_packages.
    """
    for item in os.listdir(dirname):
        if item.endswith('.dist-info') and not any(
            package in item for package in keep_packages
        ):
            shutil.rmtree(os.path.join(dirname, item))



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.clidriver import AWSCLIEntryPoint
from awscli.testutils import create_clidriver, mock
from tests.functional.history import BaseHistoryCommandParamsTest


class TestShowCommand(BaseHistoryCommandParamsTest):
    def test_show_latest(self):
        self.parsed_responses = [
            {
                "Regions": [
                    {
                        "Endpoint": "ec2.ap-south-1.amazonaws.com",
                        "RegionName": "ap-south-1",
                    },
                ]
            }
        ]
        self.run_cmd('ec2 describe-regions', expected_rc=0)
        stdout, _, _ = self.run_cmd('history show', expected_rc=0)
        # Test that the CLI specific events are present such as arguments
        # entered and version
        #
        # The show command writes the history out as binary to the attached
        # stream so we want to determine if the values are in the binary
        # stdout stream
        self.assertIn('describe-regions', stdout)
        self.assertIn('version', stdout)

    def test_show_nothing_when_no_history(self):
        self.environ['AWS_CONFIG_FILE'] = ''
        self.driver = create_clidriver()
        self.entry_point = AWSCLIEntryPoint(self.driver)
        self.parsed_responses = [
            {
                "Regions": [
                    {
                        "Endpoint": "ec2.ap-south-1.amazonaws.com",
                        "RegionName": "ap-south-1",
                    },
                ]
            }
        ]
        self.run_cmd('ec2 describe-regions', expected_rc=0)
        stdout, _, _ = self.run_cmd('history show', expected_rc=0)
        # The history show should not display anything as no history should
        # have been collected
        self.assertEqual('', stdout)

    def test_show_with_include(self):
        self.parsed_responses = [
            {
                "Regions": [
                    {
                        "Endpoint": "ec2.ap-south-1.amazonaws.com",
                        "RegionName": "ap-south-1",
                    },
                ]
            }
        ]
        self.run_cmd('ec2 describe-regions', expected_rc=0)
        stdout, _, _ = self.run_cmd(
            'history show --include CLI_ARGUMENTS', expected_rc=0
        )
        # Make sure the CLI version was not included because of the filter.
        #
        # The show command writes the history out as binary to the attached
        # stream so we want to determine if the values are in the binary
        # stdout stream
        self.assertIn('describe-regions', stdout)
        self.assertNotIn('version', stdout)

    def test_show_with_exclude(self):
        self.parsed_responses = [
            {
                "Regions": [
                    {
                        "Endpoint": "ec2.ap-south-1.amazonaws.com",
                        "RegionName": "ap-south-1",
                    },
                ]
            }
        ]
        self.run_cmd('ec2 describe-regions', expected_rc=0)
        stdout, _, _ = self.run_cmd(
            'history show --exclude CLI_ARGUMENTS', expected_rc=0
        )
        # Make sure the API call was not included because of the filter,
        # but all other events such as the version are included.
        #
        # The show command writes the history out as binary to the attached
        # stream so we want to determine if the values are in the binary
        # stdout stream
        self.assertNotIn("['ec2', 'describe-regions']", stdout)
        self.assertIn('version', stdout)



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import io
from collections import namedtuple

import pytest
from botocore.exceptions import (
    ClientError,
    NoCredentialsError,
    NoRegionError,
)
from botocore.exceptions import (
    ParamValidationError as BotocoreParamValidationError,
)

from awscli import errorhandler
from awscli.argparser import ArgParseException
from awscli.argprocess import ParamError, ParamSyntaxError
from awscli.arguments import UnknownArgumentError
from awscli.autoprompt.factory import PrompterKeyboardInterrupt
from awscli.customizations.exceptions import (
    ConfigurationError,
    ParamValidationError,
)

Case = namedtuple('Case', ['exception', 'rc', 'stderr', 'stdout'])


def _assert_rc_and_error_message(case, error_handler):
    stderr = io.StringIO()
    stdout = io.StringIO()
    try:
        raise case.exception
    except BaseException as e:
        cr = error_handler.handle_exception(e, stdout, stderr)
        assert cr == case.rc, case.exception.__class__
        assert case.stderr in stderr.getvalue()
        assert case.stdout == stdout.getvalue()


@pytest.mark.parametrize(
    "case",
    [
        Case(Exception('error'), 255, 'error', ''),
        Case(KeyboardInterrupt(), 130, '', '\n'),
        Case(NoRegionError(), 253, 'region', ''),
        Case(NoCredentialsError(), 253, 'credentials', ''),
        Case(
            ClientError(error_response={}, operation_name=''),
            254,
            'An error occurred',
            '',
        ),
        Case(
            BotocoreParamValidationError(report='param_name'),
            252,
            'param_name',
            '',
        ),
        Case(UnknownArgumentError('error'), 252, 'error', ''),
        Case(ArgParseException('error'), 252, 'error', ''),
        Case(ParamSyntaxError('error'), 252, 'error', ''),
        Case(
            ParamError(cli_name='cli', message='message'),
            252,
            "'cli': message",
            '',
        ),
        Case(ParamValidationError('error'), 252, 'error', ''),
        Case(ConfigurationError('error'), 253, 'error', ''),
    ],
)
def test_cli_error_handling_chain(case):
    error_handler = errorhandler.construct_cli_error_handlers_chain()
    _assert_rc_and_error_message(case, error_handler)


@pytest.mark.parametrize(
    "case",
    [
        Case(Exception('error'), 255, 'error', ''),
        Case(KeyboardInterrupt(), 130, '', '\n'),
        Case(NoRegionError(), 253, 'region', ''),
        Case(NoCredentialsError(), 253, 'credentials', ''),
        Case(
            ClientError(error_response={}, operation_name=''),
            254,
            'An error occurred',
            '',
        ),
        Case(BotocoreParamValidationError(report='param_name'), 252, '', ''),
        Case(UnknownArgumentError('error'), 252, '', ''),
        Case(ArgParseException('error'), 252, '', ''),
        Case(ParamSyntaxError('error'), 252, '', ''),
        Case(ParamError(cli_name='cli', message='message'), 252, '', ''),
        Case(ParamValidationError('error'), 252, '', ''),
        Case(ConfigurationError('error'), 253, 'error', ''),
    ],
)
def test_cli_error_handling_chain_injection(case):
    error_handler = errorhandler.construct_cli_error_handlers_chain()
    error_handler.inject_handler(
        0, errorhandler.SilenceParamValidationMsgErrorHandler()
    )
    _assert_rc_and_error_message(case, error_handler)


@pytest.mark.parametrize(
    "case",
    [
        Case(Exception('error'), 255, 'error', ''),
        Case(KeyboardInterrupt(), 130, '', '\n'),
        Case(PrompterKeyboardInterrupt('error'), 130, 'error', ''),
        Case(ParamValidationError('error'), 252, 'error', ''),
    ],
)
def test_entry_point_error_handling_chain(case):
    error_handler = errorhandler.construct_entry_point_handlers_chain()
    _assert_rc_and_error_message(case, error_handler)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.session import Session

from awscli.customizations.exceptions import ParamValidationError
from awscli.customizations.wizard.app import WizardAppRunner
from awscli.customizations.wizard.commands import TopLevelWizardCommand
from awscli.customizations.wizard.core import Runner
from awscli.customizations.wizard.loader import WizardLoader
from awscli.testutils import mock, unittest


class TestWizardCommand(unittest.TestCase):
    def setUp(self):
        self.session = mock.Mock(spec=Session)
        self.session.emit_first_non_none_response.return_value = None
        self.loader = mock.Mock(spec=WizardLoader)
        self.loader.list_available_wizards.return_value = ['mywizard']
        self.wizard_v2_runner = mock.Mock(spec=WizardAppRunner)
        self.wizard_v1_runner = mock.Mock(spec=Runner)
        self.runner = {
            '0.1': self.wizard_v1_runner,
            '0.2': self.wizard_v2_runner,
        }

    def test_will_delegate_v1_definition_to_v1_runner(self):
        loaded = {'version': '0.1'}
        self.loader.load_wizard.return_value = loaded
        cmd = TopLevelWizardCommand(
            self.session,
            loader=self.loader,
            parent_command='',
            runner=self.runner,
        )
        cmd(['mywizard'], parsed_globals=None)
        self.wizard_v1_runner.run.assert_called_with(loaded)

    def test_will_delegate_v2_definition_to_v2_runner(self):
        loaded = {'version': '0.2'}
        self.loader.load_wizard.return_value = loaded
        cmd = TopLevelWizardCommand(
            self.session,
            loader=self.loader,
            parent_command='',
            runner=self.runner,
        )
        cmd(['mywizard'], parsed_globals=None)
        self.wizard_v2_runner.run.assert_called_with(loaded)

    def test_will_raise_exception_for_unsupported_version(self):
        loaded = {'version': '100'}
        self.loader.load_wizard.return_value = loaded
        cmd = TopLevelWizardCommand(
            self.session,
            loader=self.loader,
            parent_command='',
            runner=self.runner,
        )
        with self.assertRaises(ParamValidationError) as e:
            cmd(['mywizard'], parsed_globals=None)
        self.assertIn('file has unsupported version', str(e.exception))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import glob
import os
import shutil
import tempfile

from botocore.compat import OrderedDict

from awscli.customizations.eks.exceptions import EKSError
from awscli.customizations.eks.kubeconfig import (
    Kubeconfig,
    KubeconfigAppender,
    KubeconfigCorruptedError,
    KubeconfigError,
    KubeconfigInaccessableError,
    KubeconfigLoader,
    KubeconfigValidator,
    KubeconfigWriter,
    _get_new_kubeconfig_content,
)
from awscli.customizations.eks.ordered_yaml import ordered_yaml_load
from awscli.customizations.utils import uni_print
from awscli.testutils import skip_if_windows, unittest
from tests.functional.eks.test_util import get_testdata


class TestKubeconfig(unittest.TestCase):
    def setUp(self):
        self._content = OrderedDict([("apiVersion", "v1")])
        self._path = "/some_path"

    def test_no_content(self):
        config = Kubeconfig(self._path, None)
        self.assertEqual(config.content, _get_new_kubeconfig_content())

    def test_has_cluster(self):
        self._content["clusters"] = [
            OrderedDict([("cluster", None), ("name", "clustername")]),
            OrderedDict([("cluster", None), ("name", "anotherclustername")]),
        ]

        config = Kubeconfig(self._path, self._content)
        self.assertTrue(config.has_cluster("clustername"))
        self.assertFalse(config.has_cluster("othercluster"))

    def test_has_cluster_with_no_clusters(self):
        config = Kubeconfig(self._path, self._content)
        self.assertFalse(config.has_cluster("clustername"))

    def test_has_cluster_with_none_clusters(self):
        self._content["clusters"] = None

        config = Kubeconfig(self._path, self._content)
        self.assertFalse(config.has_cluster("anyclustername"))

    def test_has_cluster_with_cluster_no_name(self):
        self._content["clusters"] = [
            OrderedDict([("cluster", None), ("name", "clustername")]),
            OrderedDict(
                [
                    ("cluster", None),
                    ("name", None),
                ]
            ),
            OrderedDict(
                [
                    ("cluster", None),
                ]
            ),
        ]

        config = Kubeconfig(self._path, self._content)
        self.assertTrue(config.has_cluster("clustername"))
        self.assertFalse(config.has_cluster("othercluster"))


class TestKubeconfigWriter(unittest.TestCase):
    @skip_if_windows('Read permissions tests only supported on mac/linux')
    def test_not_world_readable(self):
        tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tmpdir)
        config_path = os.path.join(tmpdir, "config")
        config = Kubeconfig(config_path, None)
        KubeconfigWriter().write_kubeconfig(config)
        stat = os.stat(config_path)
        self.assertEqual(stat.st_mode & 0o777, 0o600)

    def test_truncates(self):
        tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tmpdir)
        config_path = os.path.join(tmpdir, "config")
        with open(config_path, "w+") as f:
            f.write("#" * 100)
        KubeconfigWriter().write_kubeconfig(Kubeconfig(config_path, {}))
        empty_stat = os.stat(config_path)
        self.assertLessEqual(
            empty_stat.st_size,
            4,
            "file should be '{}[newline]', 3/4 bytes long ",
        )


class TestKubeconfigValidator(unittest.TestCase):
    def setUp(self):
        self._validator = KubeconfigValidator()

    def test_valid(self):
        valid_cases = glob.glob(get_testdata("valid_*"))
        for case in valid_cases:
            with open(case) as stream:
                content_dict = ordered_yaml_load(stream)
            if content_dict is not None:
                config = Kubeconfig(None, content_dict)
                try:
                    self._validator.validate_config(config)
                except KubeconfigError as e:
                    self.fail(f"Valid file {case} raised {e}.")

    def test_invalid(self):
        invalid_cases = glob.glob(get_testdata("invalid_*"))
        for case in invalid_cases:
            with open(case) as stream:
                content_dict = ordered_yaml_load(stream)
            config = Kubeconfig(None, content_dict)
            self.assertRaises(
                KubeconfigCorruptedError,
                self._validator.validate_config,
                config,
            )


class TestKubeconfigAppender(unittest.TestCase):
    def setUp(self):
        self._appender = KubeconfigAppender()

    def test_basic_insert(self):
        initial = OrderedDict(
            [
                ("apiVersion", "v1"),
                (
                    "clusters",
                    [
                        OrderedDict(
                            [
                                (
                                    "cluster",
                                    OrderedDict(
                                        [
                                            (
                                                "certificate-authority-data",
                                                "data1",
                                            ),
                                            ("server", "endpoint1"),
                                        ]
                                    ),
                                ),
                                ("name", "oldclustername"),
                            ]
                        )
                    ],
                ),
                ("contexts", []),
                ("current-context", "simple"),
                ("kind", "Config"),
                ("preferences", OrderedDict()),
                ("users", []),
            ]
        )
        cluster = OrderedDict(
            [
                (
                    "cluster",
                    OrderedDict(
                        [
                            ("certificate-authority-data", "data2"),
                            ("server", "endpoint2"),
                        ]
                    ),
                ),
                ("name", "clustername"),
            ]
        )
        cluster_added_correct = OrderedDict(
            [
                ("apiVersion", "v1"),
                (
                    "clusters",
                    [
                        OrderedDict(
                            [
                                (
                                    "cluster",
                                    OrderedDict(
                                        [
                                            (
                                                "certificate-authority-data",
                                                "data1",
                                            ),
                                            ("server", "endpoint1"),
                                        ]
                                    ),
                                ),
                                ("name", "oldclustername"),
                            ]
                        ),
                        OrderedDict(
                            [
                                (
                                    "cluster",
                                    OrderedDict(
                                        [
                                            (
                                                "certificate-authority-data",
                                                "data2",
                                            ),
                                            ("server", "endpoint2"),
                                        ]
                                    ),
                                ),
                                ("name", "clustername"),
                            ]
                        ),
                    ],
                ),
                ("contexts", []),
                ("current-context", "simple"),
                ("kind", "Config"),
                ("preferences", OrderedDict()),
                ("users", []),
            ]
        )
        cluster_added = self._appender.insert_entry(
            Kubeconfig(None, initial), "clusters", cluster
        )
        self.assertDictEqual(cluster_added.content, cluster_added_correct)

    def test_update_existing(self):
        initial = OrderedDict(
            [
                ("apiVersion", "v1"),
                (
                    "clusters",
                    [
                        OrderedDict(
                            [
                                (
                                    "cluster",
                                    OrderedDict([("server", "endpoint")]),
                                ),
                                ("name", "clustername"),
                            ]
                        )
                    ],
                ),
                ("contexts", []),
                ("current-context", None),
                ("kind", "Config"),
                ("preferences", OrderedDict()),
                ("users", []),
            ]
        )
        cluster = OrderedDict(
            [
                (
                    "cluster",
                    OrderedDict(
                        [
                            ("certificate-authority-data", "data"),
                            ("server", "endpoint"),
                        ]
                    ),
                ),
                ("name", "clustername"),
            ]
        )
        correct = OrderedDict(
            [
                ("apiVersion", "v1"),
                (
                    "clusters",
                    [
                        OrderedDict(
                            [
                                (
                                    "cluster",
                                    OrderedDict(
                                        [
                                            (
                                                "certificate-authority-data",
                                                "data",
                                            ),
                                            ("server", "endpoint"),
                                        ]
                                    ),
                                ),
                                ("name", "clustername"),
                            ]
                        )
                    ],
                ),
                ("contexts", []),
                ("current-context", None),
                ("kind", "Config"),
                ("preferences", OrderedDict()),
                ("users", []),
            ]
        )
        updated = self._appender.insert_entry(
            Kubeconfig(None, initial), "clusters", cluster
        )
        self.assertDictEqual(updated.content, correct)

    def test_key_not_exist(self):
        initial = OrderedDict(
            [
                ("apiVersion", "v1"),
                ("contexts", []),
                ("current-context", None),
                ("kind", "Config"),
                ("preferences", OrderedDict()),
                ("users", []),
            ]
        )
        cluster = OrderedDict(
            [
                (
                    "cluster",
                    OrderedDict(
                        [
                            ("certificate-authority-data", "data"),
                            ("server", "endpoint"),
                        ]
                    ),
                ),
                ("name", "clustername"),
            ]
        )
        correct = OrderedDict(
            [
                ("apiVersion", "v1"),
                ("contexts", []),
                ("current-context", None),
                ("kind", "Config"),
                ("preferences", OrderedDict()),
                ("users", []),
                (
                    "clusters",
                    [
                        OrderedDict(
                            [
                                (
                                    "cluster",
                                    OrderedDict(
                                        [
                                            (
                                                "certificate-authority-data",
                                                "data",
                                            ),
                                            ("server", "endpoint"),
                                        ]
                                    ),
                                ),
                                ("name", "clustername"),
                            ]
                        )
                    ],
                ),
            ]
        )
        updated = self._appender.insert_entry(
            Kubeconfig(None, initial), "clusters", cluster
        )
        self.assertDictEqual(updated.content, correct)

    def test_key_none(self):
        initial = OrderedDict(
            [
                ("apiVersion", "v1"),
                ("clusters", None),
                ("contexts", []),
                ("current-context", None),
                ("kind", "Config"),
                ("preferences", OrderedDict()),
                ("users", []),
            ]
        )
        cluster = OrderedDict(
            [
                (
                    "cluster",
                    OrderedDict(
                        [
                            ("certificate-authority-data", "data"),
                            ("server", "endpoint"),
                        ]
                    ),
                ),
                ("name", "clustername"),
            ]
        )
        correct = OrderedDict(
            [
                ("apiVersion", "v1"),
                (
                    "clusters",
                    [
                        OrderedDict(
                            [
                                (
                                    "cluster",
                                    OrderedDict(
                                        [
                                            (
                                                "certificate-authority-data",
                                                "data",
                                            ),
                                            ("server", "endpoint"),
                                        ]
                                    ),
                                ),
                                ("name", "clustername"),
                            ]
                        )
                    ],
                ),
                ("contexts", []),
                ("current-context", None),
                ("kind", "Config"),
                ("preferences", OrderedDict()),
                ("users", []),
            ]
        )
        updated = self._appender.insert_entry(
            Kubeconfig(None, initial), "clusters", cluster
        )
        self.assertDictEqual(updated.content, correct)

    def test_key_not_array(self):
        initial = OrderedDict(
            [
                ("apiVersion", "v1"),
                ("contexts", []),
                ("current-context", None),
                ("kind", "Config"),
                ("preferences", OrderedDict()),
                ("users", []),
            ]
        )
        cluster = OrderedDict(
            [
                (
                    "cluster",
                    OrderedDict(
                        [
                            ("certificate-authority-data", "data"),
                            ("server", "endpoint"),
                        ]
                    ),
                ),
                ("name", "clustername"),
            ]
        )
        self.assertRaises(
            KubeconfigError,
            self._appender.insert_entry,
            Kubeconfig(None, initial),
            "kind",
            cluster,
        )

    def test_make_context(self):
        cluster = OrderedDict(
            [("name", "clustername"), ("cluster", OrderedDict())]
        )
        user = OrderedDict([("name", "username"), ("user", OrderedDict())])
        context_correct = OrderedDict(
            [
                (
                    "context",
                    OrderedDict(
                        [("cluster", "clustername"), ("user", "username")]
                    ),
                ),
                ("name", "username"),
            ]
        )
        context = self._appender._make_context(cluster, user)
        self.assertDictEqual(context, context_correct)

    def test_make_context_alias(self):
        cluster = OrderedDict(
            [("name", "clustername"), ("cluster", OrderedDict())]
        )
        user = OrderedDict([("name", "username"), ("user", OrderedDict())])
        context_correct = OrderedDict(
            [
                (
                    "context",
                    OrderedDict(
                        [("cluster", "clustername"), ("user", "username")]
                    ),
                ),
                ("name", "alias"),
            ]
        )
        alias = "alias"
        context = self._appender._make_context(cluster, user, alias=alias)
        self.assertDictEqual(context, context_correct)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from copy import deepcopy

from botocore.model import ServiceModel

from awscli.autocomplete.autogen import ServerCompletionHeuristic
from awscli.testutils import unittest

# To make these tests more realistic, a stripped down version of the ACM
# service is included as part of these tests.  There are a few extra
# shapes/operations that are added to exercise more of the code, but
# overall the model still reflects something that you'd actually see
# in a real service model.
BASIC_MODEL = {
    "version": "2.0",
    "metadata": {
        "apiVersion": "2015-12-08",
        "endpointPrefix": "acm",
        "jsonVersion": "1.1",
        "protocol": "json",
        "serviceAbbreviation": "ACM",
        "serviceFullName": "AWS Certificate Manager",
        "serviceId": "ACM",
        "signatureVersion": "v4",
        "targetPrefix": "CertificateManager",
        "uid": "acm-2015-12-08",
    },
    "operations": {
        "CreateCertificate": {
            "http": {"method": "POST", "requestUri": "/"},
            "input": {"shape": "CreateCertificateRequest"},
            "name": "DeleteCertificate",
        },
        "DeleteCertificate": {
            "http": {"method": "POST", "requestUri": "/"},
            "input": {"shape": "DeleteCertificateRequest"},
            "name": "DeleteCertificate",
        },
        "DescribeCertificate": {
            "http": {"method": "POST", "requestUri": "/"},
            "input": {"shape": "DescribeCertificateRequest"},
            "name": "DescribeCertificate",
            "output": {"shape": "DescribeCertificateResponse"},
        },
        "ExportCertificate": {
            "http": {"method": "POST", "requestUri": "/"},
            "input": {"shape": "ExportCertificateRequest"},
            "name": "ExportCertificate",
            "output": {"shape": "ExportCertificateResponse"},
        },
        "GetCertificate": {
            "http": {"method": "POST", "requestUri": "/"},
            "input": {"shape": "GetCertificateRequest"},
            "name": "GetCertificate",
            "output": {"shape": "GetCertificateResponse"},
        },
        "ImportCertificate": {
            "http": {"method": "POST", "requestUri": "/"},
            "input": {"shape": "ImportCertificateRequest"},
            "name": "ImportCertificate",
            "output": {"shape": "ImportCertificateResponse"},
        },
        "ListCertificates": {
            "http": {"method": "POST", "requestUri": "/"},
            "input": {"shape": "ListCertificatesRequest"},
            "name": "ListCertificates",
            "output": {"shape": "ListCertificatesResponse"},
        },
    },
    "shapes": {
        "Arn": {"max": 2048, "min": 20, "type": "string"},
        "CertificateBody": {"max": 32768, "min": 1, "type": "string"},
        "CertificateBodyBlob": {"max": 32768, "min": 1, "type": "blob"},
        "CertificateChain": {"max": 2097152, "min": 1, "type": "string"},
        "CertificateChainBlob": {"max": 2097152, "min": 1, "type": "blob"},
        "CertificateDetail": {
            "members": {
                "CertificateArn": {"shape": "Arn"},
                "CertificateAuthorityArn": {"shape": "Arn"},
                "Options": {"shape": "CertificateOptions"},
                "SubjectAlternativeNames": {"shape": "DomainList"},
                "Type": {"shape": "CertificateType"},
            },
            "type": "structure",
        },
        "CertificateOptions": {
            "members": {},
            "type": "structure",
        },
        "CertificateStatus": {"type": "string"},
        "CertificateStatuses": {
            "member": {"shape": "CertificateStatus"},
            "type": "list",
        },
        "CertificateSummary": {
            "members": {
                "CertificateArn": {"shape": "Arn"},
                "DomainName": {"shape": "DomainNameString"},
            },
            "type": "structure",
        },
        "CertificateSummaryList": {
            "member": {"shape": "CertificateSummary"},
            "type": "list",
        },
        "CertificateType": {"type": "string"},
        "CreateCertificateRequest": {
            "members": {"CertificateArn": {"shape": "Arn"}},
            "required": ["CertificateArn"],
            "type": "structure",
        },
        "DeleteCertificateRequest": {
            "members": {"CertificateArn": {"shape": "Arn"}},
            "required": ["CertificateArn"],
            "type": "structure",
        },
        "DescribeCertificateRequest": {
            "members": {"CertificateArn": {"shape": "Arn"}},
            "required": ["CertificateArn"],
            "type": "structure",
        },
        "DescribeCertificateResponse": {
            "members": {"Certificate": {"shape": "CertificateDetail"}},
            "type": "structure",
        },
        "DomainList": {
            "max": 100,
            "member": {"shape": "DomainNameString"},
            "min": 1,
            "type": "list",
        },
        "DomainNameString": {"max": 253, "min": 1, "type": "string"},
        "DomainStatus": {"type": "string"},
        "DomainValidation": {
            "members": {
                "DomainName": {"shape": "DomainNameString"},
                "ResourceRecord": {"shape": "ResourceRecord"},
                "ValidationDomain": {"shape": "DomainNameString"},
                "ValidationEmails": {"shape": "ValidationEmailList"},
                "ValidationMethod": {"shape": "ValidationMethod"},
                "ValidationStatus": {"shape": "DomainStatus"},
            },
            "required": ["DomainName"],
            "type": "structure",
        },
        "DomainValidationList": {
            "max": 1000,
            "member": {"shape": "DomainValidation"},
            "min": 1,
            "type": "list",
        },
        "ExportCertificateRequest": {
            "members": {
                "CertificateArn": {"shape": "Arn"},
                "Passphrase": {"shape": "PassphraseBlob"},
            },
            "required": ["CertificateArn", "Passphrase"],
            "type": "structure",
        },
        "ExportCertificateResponse": {
            "members": {
                "Certificate": {"shape": "CertificateBody"},
                "CertificateChain": {"shape": "CertificateChain"},
                "PrivateKey": {"shape": "PrivateKey"},
            },
            "type": "structure",
        },
        "ExtendedKeyUsage": {
            "members": {
                "Name": {"shape": "ExtendedKeyUsageName"},
                "OID": {"shape": "String"},
            },
            "type": "structure",
        },
        "ExtendedKeyUsageFilterList": {
            "member": {"shape": "ExtendedKeyUsageName"},
            "type": "list",
        },
        "ExtendedKeyUsageList": {
            "member": {"shape": "ExtendedKeyUsage"},
            "type": "list",
        },
        "ExtendedKeyUsageName": {"type": "string"},
        "FailureReason": {"type": "string"},
        "Filters": {
            "members": {
                "extendedKeyUsage": {"shape": "ExtendedKeyUsageFilterList"},
                "keyTypes": {"shape": "KeyAlgorithmList"},
                "keyUsage": {"shape": "KeyUsageFilterList"},
            },
            "type": "structure",
        },
        "GetCertificateRequest": {
            "members": {"CertificateArn": {"shape": "Arn"}},
            "required": ["CertificateArn"],
            "type": "structure",
        },
        "GetCertificateResponse": {
            "members": {
                "Certificate": {"shape": "CertificateBody"},
                "CertificateChain": {"shape": "CertificateChain"},
            },
            "type": "structure",
        },
        "ImportCertificateRequest": {
            "members": {
                "Certificate": {"shape": "CertificateBodyBlob"},
                "CertificateArn": {"shape": "Arn"},
                "CertificateChain": {"shape": "CertificateChainBlob"},
                "PrivateKey": {"shape": "PrivateKeyBlob"},
            },
            "required": ["Certificate", "PrivateKey"],
            "type": "structure",
        },
        "ImportCertificateResponse": {
            "members": {"CertificateArn": {"shape": "Arn"}},
            "type": "structure",
        },
        "InUseList": {"member": {"shape": "String"}, "type": "list"},
        "KeyAlgorithm": {"type": "string"},
        "KeyAlgorithmList": {
            "member": {"shape": "KeyAlgorithm"},
            "type": "list",
        },
        "KeyUsage": {
            "members": {"Name": {"shape": "KeyUsageName"}},
            "type": "structure",
        },
        "KeyUsageFilterList": {
            "member": {"shape": "KeyUsageName"},
            "type": "list",
        },
        "KeyUsageList": {"member": {"shape": "KeyUsage"}, "type": "list"},
        "KeyUsageName": {"type": "string"},
        "ListCertificatesRequest": {
            "members": {
                "CertificateStatuses": {"shape": "CertificateStatuses"},
                "Includes": {"shape": "Filters"},
                "MaxItems": {"shape": "MaxItems"},
                "NextToken": {"shape": "NextToken"},
            },
            "type": "structure",
        },
        "ListCertificatesResponse": {
            "members": {
                "CertificateSummaryList": {"shape": "CertificateSummaryList"},
                "NextToken": {"shape": "NextToken"},
            },
            "type": "structure",
        },
        "MaxItems": {"max": 1000, "min": 1, "type": "integer"},
        "NextToken": {"max": 320, "min": 1, "type": "string"},
        "PassphraseBlob": {
            "type": "blob",
        },
        "PrivateKey": {
            "type": "string",
        },
        "PrivateKeyBlob": {
            "type": "blob",
        },
        "RecordType": {"type": "string"},
        "RenewalEligibility": {"type": "string"},
        "RenewalStatus": {"type": "string"},
        "RenewalSummary": {
            "members": {
                "DomainValidationOptions": {"shape": "DomainValidationList"},
                "RenewalStatus": {"shape": "RenewalStatus"},
            },
            "required": ["RenewalStatus", "DomainValidationOptions"],
            "type": "structure",
        },
        "ResourceRecord": {
            "members": {
                "Name": {"shape": "String"},
                "Type": {"shape": "RecordType"},
                "Value": {"shape": "String"},
            },
            "required": ["Name", "Type", "Value"],
            "type": "structure",
        },
        "RevocationReason": {"type": "string"},
        "String": {"type": "string"},
        "TStamp": {"type": "timestamp"},
        "ValidationEmailList": {"member": {"shape": "String"}, "type": "list"},
        "ValidationMethod": {"type": "string"},
    },
}


# This is a stripped down version from athena.
MODEL_WITH_STRING_LIST = {
    "version": "2.0",
    "metadata": {},
    "operations": {
        "CreateNamedQuery": {
            "http": {"method": "POST", "requestUri": "/"},
            "input": {"shape": "CreateNamedQueryInput"},
            "name": "CreateNamedQuery",
            "output": {"shape": "CreateNamedQueryOutput"},
        },
        "DeleteNamedQuery": {
            "http": {"method": "POST", "requestUri": "/"},
            "input": {"shape": "DeleteNamedQueryInput"},
            "name": "DeleteNamedQuery",
            "output": {"shape": "DeleteNamedQueryOutput"},
        },
        "GetNamedQuery": {
            "http": {"method": "POST", "requestUri": "/"},
            "input": {"shape": "GetNamedQueryInput"},
            "name": "GetNamedQuery",
            "output": {"shape": "GetNamedQueryOutput"},
        },
        "ListNamedQueries": {
            "http": {"method": "POST", "requestUri": "/"},
            "input": {"shape": "ListNamedQueriesInput"},
            "name": "ListNamedQueries",
            "output": {"shape": "ListNamedQueriesOutput"},
        },
        "ListStrangeStuffAndNamedQueries": {
            "http": {"method": "POST", "requestUri": "/"},
            "input": {"shape": "ListNamedQueriesInput"},
            "name": "ListStrangeStuffAndNamedQueries",
            "output": {"shape": "ListNamedQueriesOutput"},
        },
    },
    "shapes": {
        "CreateNamedQueryInput": {
            "members": {
                "ClientRequestToken": {
                    "idempotencyToken": True,
                    "shape": "IdempotencyToken",
                },
                "Database": {"shape": "DatabaseString"},
                "Description": {"shape": "DescriptionString"},
                "Name": {"shape": "NameString"},
                "QueryString": {"shape": "QueryString"},
            },
            "required": ["Name", "Database", "QueryString"],
            "type": "structure",
        },
        "CreateNamedQueryOutput": {
            "members": {"NamedQueryId": {"shape": "NamedQueryId"}},
            "type": "structure",
        },
        "DatabaseString": {"type": "string"},
        "DeleteNamedQueryInput": {
            "members": {
                "NamedQueryId": {
                    "idempotencyToken": True,
                    "shape": "NamedQueryId",
                }
            },
            "required": ["NamedQueryId"],
            "type": "structure",
        },
        "DeleteNamedQueryOutput": {"members": {}, "type": "structure"},
        "DescriptionString": {"type": "string"},
        "GetNamedQueryInput": {
            "members": {"NamedQueryId": {"shape": "NamedQueryId"}},
            "required": ["NamedQueryId"],
            "type": "structure",
        },
        "GetNamedQueryOutput": {
            "members": {"NamedQuery": {"shape": "NamedQuery"}},
            "type": "structure",
        },
        "IdempotencyToken": {"type": "string"},
        "ListNamedQueriesInput": {
            "members": {
                "MaxResults": {"shape": "MaxNamedQueriesCount"},
                "NextToken": {"shape": "Token"},
            },
            "type": "structure",
        },
        "ListNamedQueriesOutput": {
            "members": {
                "NamedQueryIds": {"shape": "NamedQueryIdList"},
                "NextToken": {"shape": "Token"},
            },
            "type": "structure",
        },
        "MaxNamedQueriesCount": {"type": "integer"},
        "NameString": {"type": "string"},
        "NamedQuery": {
            "members": {
                "Database": {"shape": "DatabaseString"},
                "Description": {"shape": "DescriptionString"},
                "Name": {"shape": "NameString"},
                "NamedQueryId": {"shape": "NamedQueryId"},
                "QueryString": {"shape": "QueryString"},
            },
            "required": ["Name", "Database", "QueryString"],
            "type": "structure",
        },
        "NamedQueryId": {"type": "string"},
        "NamedQueryIdList": {
            "member": {"shape": "NamedQueryId"},
            "type": "list",
        },
        "QueryString": {"type": "string"},
        "Token": {"type": "string"},
    },
}


class TestCanGenerateCompletions(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        self.service_model = ServiceModel(BASIC_MODEL)
        self.heuristic = ServerCompletionHeuristic()

    def test_can_generate_resource_descriptions(self):
        resources = self.heuristic.generate_completion_descriptions(
            self.service_model
        )['resources']
        self.assertEqual(
            resources,
            {
                'Certificate': {
                    'operation': 'ListCertificates',
                    'resourceIdentifier': {
                        'CertificateArn': (
                            'CertificateSummaryList[].CertificateArn'
                        )
                    },
                }
            },
        )

    def test_can_generate_operations(self):
        # This model is chosen specifically because there's only
        # one resource we generate.  This makes the assertions easier.
        operations = self.heuristic.generate_completion_descriptions(
            self.service_model
        )['operations']
        completion_param = {
            'CertificateArn': {
                'completions': [
                    {
                        'parameters': {},
                        'resourceIdentifier': 'CertificateArn',
                        'resourceName': 'Certificate',
                    }
                ]
            }
        }
        self.assertEqual(
            list(sorted(operations)),
            [
                'DeleteCertificate',
                'DescribeCertificate',
                'ExportCertificate',
                'GetCertificate',
                'ImportCertificate',
            ],
        )
        for op in operations.values():
            self.assertEqual(op, completion_param)

    def test_can_generate_operation_with_describe_prefix(self):
        model = deepcopy(BASIC_MODEL)
        # Swap ListCertificates for DescribeCertificates.
        # We should still be able to generate auto-completion data.
        model['operations']['DescribeCertificates'] = model['operations'].pop(
            'ListCertificates'
        )
        service_model = ServiceModel(model)
        completion_data = self.heuristic.generate_completion_descriptions(
            service_model
        )
        # Ensure we're using the swapped 'DescribeCertificates' operation.
        self.assertEqual(
            completion_data['resources']['Certificate']['operation'],
            'DescribeCertificates',
        )
        self.assertEqual(
            list(sorted(completion_data['operations'])),
            [
                'DeleteCertificate',
                'DescribeCertificate',
                'ExportCertificate',
                'GetCertificate',
                'ImportCertificate',
            ],
        )

    def test_can_generate_string_list_completions(self):
        service_model = ServiceModel(MODEL_WITH_STRING_LIST)
        completion_data = self.heuristic.generate_completion_descriptions(
            service_model
        )
        expected = {
            'version': '1.0',
            'operations': {
                'DeleteNamedQuery': {
                    'NamedQueryId': {
                        'completions': [
                            {
                                'parameters': {},
                                'resourceIdentifier': 'NamedQueryId',
                                'resourceName': 'NamedQuery',
                            }
                        ]
                    }
                },
                'GetNamedQuery': {
                    'NamedQueryId': {
                        'completions': [
                            {
                                'parameters': {},
                                'resourceIdentifier': 'NamedQueryId',
                                'resourceName': 'NamedQuery',
                            }
                        ]
                    }
                },
            },
            'resources': {
                'NamedQuery': {
                    'operation': 'ListNamedQueries',
                    'resourceIdentifier': {'NamedQueryId': 'NamedQueryIds[]'},
                }
            },
        }
        self.assertEqual(completion_data, expected)

    def test_resources_map_input_params_for_required_inputs(self):
        model_dict = deepcopy(BASIC_MODEL)
        # We're going to mark an input param of ListCertificates as required.
        model_dict['shapes']['ListCertificatesRequest']['required'] = [
            'CertificateStatuses'
        ]
        # We also have to have an operation reference this
        service_model = ServiceModel(model_dict)
        completion_data = self.heuristic.generate_completion_descriptions(
            service_model, prune_completions=False
        )
        resources = completion_data['resources']
        self.assertEqual(
            resources,
            {
                'Certificate': {
                    'operation': 'ListCertificates',
                    'inputParameters': ['CertificateStatuses'],
                    'resourceIdentifier': {
                        'CertificateArn': (
                            'CertificateSummaryList[].CertificateArn'
                        ),
                        'DomainName': ('CertificateSummaryList[].DomainName'),
                    },
                }
            },
        )

    def test_remove_operations_with_required_params(self):
        # We can remove this test once we implement this functionality.
        custom_model = {
            'metadata': {},
            'operations': {
                'ListFooBarThings': {
                    'input': {'shape': 'ListFooBarThingsRequest'},
                    'output': {'shape': 'ListFooBarThingsResponse'},
                },
                'DeleteFooBarThing': {
                    'input': {'shape': 'DeleteFooBarThingRequest'},
                },
            },
            'shapes': {
                'DeleteFooBarThingRequest': {
                    'type': 'structure',
                    'members': {
                        'RequiredParam': {'shape': 'String'},
                        'FooBarThing': {'shape': 'String'},
                    },
                },
                'ListFooBarThingsRequest': {
                    'members': {
                        'RequiredParam': {'shape': 'String'},
                    },
                    'type': 'structure',
                    'required': ['RequiredParam'],
                },
                'ListFooBarThingsResponse': {
                    'type': 'structure',
                    'members': {
                        'FooBarThings': {'shape': 'FooBarThingList'},
                    },
                },
                'FooBarThingList': {
                    'type': 'list',
                    'member': {'shape': 'String'},
                },
                'String': {'type': 'string'},
            },
        }
        service_model = ServiceModel(custom_model)
        completion_data = self.heuristic.generate_completion_descriptions(
            service_model
        )
        # The operations dict should be empty because the FooBarThing has
        # a required parameter and we don't support that yet.
        self.assertEqual(completion_data['operations'], {})

    def test_can_reference_multiple_identifiers_if_used(self):
        custom_model = {
            'metadata': {},
            'operations': {
                'ListFooBarThings': {
                    'output': {'shape': 'ListFooBarThingsResponse'},
                },
                'DeleteFooBarThing': {
                    'input': {'shape': 'DeleteFooBarThingRequest'},
                },
            },
            'shapes': {
                'DeleteFooBarThingRequest': {
                    'type': 'structure',
                    'members': {
                        'FooBarThingId': {'shape': 'String'},
                        'FooBarThingArn': {'shape': 'String'},
                    },
                },
                'ListFooBarThingsResponse': {
                    'type': 'structure',
                    'members': {
                        'FooBarThings': {'shape': 'FooBarThingList'},
                    },
                },
                'FooBarThingList': {
                    'type': 'list',
                    'member': {'shape': 'FooBarThingType'},
                },
                'FooBarThingType': {
                    'type': 'structure',
                    'members': {
                        # The DeleteFooBarThing accepts either an Id
                        # or an Arn.
                        'FooBarThingId': {'shape': 'String'},
                        'FooBarThingArn': {'shape': 'String'},
                        # However it doesn't accept a "Name", so this
                        # identifier will be pruned from the response.
                        'FooBarThingPruneId': {'shape': 'String'},
                    },
                },
                'String': {'type': 'string'},
            },
        }
        service_model = ServiceModel(custom_model)
        completion_data = self.heuristic.generate_completion_descriptions(
            service_model
        )
        self.assertEqual(
            completion_data['resources'],
            {
                'FooBarThing': {
                    'operation': 'ListFooBarThings',
                    'resourceIdentifier': {
                        'FooBarThingArn': 'FooBarThings[].FooBarThingArn',
                        'FooBarThingId': 'FooBarThings[].FooBarThingId',
                    },
                }
            },
        )



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import time

from botocore.exceptions import HTTPClientError
from s3transfer.exceptions import CancelledError, FatalError

from awscli.compat import StringIO, queue
from awscli.customizations.s3.results import (
    CommandResultRecorder,
    CtrlCResult,
    DoneResultSubscriber,
    DryRunResult,
    ErrorResult,
    FailureResult,
    FinalTotalSubmissionsResult,
    NoProgressResultPrinter,
    OnlyShowErrorsResultPrinter,
    ProgressResult,
    ProgressResultSubscriber,
    QueuedResult,
    QueuedResultSubscriber,
    ResultPrinter,
    ResultProcessor,
    ResultRecorder,
    ShutdownThreadRequest,
    SkipFileResult,
    SuccessResult,
)
from awscli.customizations.s3.utils import WarningResult
from awscli.testutils import mock, unittest
from tests.unit.customizations.s3 import (
    FakeTransferFuture,
    FakeTransferFutureCallArgs,
    FakeTransferFutureMeta,
)


class TestResultSubscribers(unittest.TestCase):
    def setUp(self):
        self.result_queue = queue.Queue()

        self.bucket = 'mybucket'
        self.key = 'mykey'
        self.filename = 'myfile'
        self.size = 20 * (1024 * 1024)  # 20 MB

        self.ref_exception = Exception()
        self.set_ref_transfer_futures()

        self.src = 'src'
        self.dest = 'dest'
        self.transfer_type = 'upload'

    def set_ref_transfer_futures(self):
        self.future = self.get_success_transfer_future('foo')
        self.failure_future = self.get_failed_transfer_future(
            self.ref_exception
        )

    def get_success_transfer_future(self, result):
        return self._get_transfer_future(result=result)

    def get_failed_transfer_future(self, exception):
        return self._get_transfer_future(exception=exception)

    def _get_transfer_future(self, result=None, exception=None):
        call_args = self._get_transfer_future_call_args()
        meta = FakeTransferFutureMeta(size=self.size, call_args=call_args)
        return FakeTransferFuture(
            result=result, exception=exception, meta=meta
        )

    def _get_transfer_future_call_args(self):
        return FakeTransferFutureCallArgs(
            fileobj=self.filename, key=self.key, bucket=self.bucket
        )

    def get_queued_result(self):
        return self.result_queue.get(block=False)

    def assert_result_queue_is_empty(self):
        self.assertTrue(self.result_queue.empty())

    def get_result_subscriber(self, subscriber_cls):
        return subscriber_cls(
            result_queue=self.result_queue,
            transfer_type=self.transfer_type,
            src=self.src,
            dest=self.dest,
            size=self.size,
        )

    def test_on_queued(self):
        subscriber = self.get_result_subscriber(QueuedResultSubscriber)
        subscriber.on_queued(self.future)
        result = self.get_queued_result()
        self.assert_result_queue_is_empty()
        self.assertEqual(
            result,
            QueuedResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                total_transfer_size=self.size,
            ),
        )

    def test_on_progress(self):
        ref_bytes_transferred = 1024 * 1024  # 1MB
        subscriber = self.get_result_subscriber(ProgressResultSubscriber)
        subscriber.on_progress(self.future, ref_bytes_transferred)
        result = self.get_queued_result()
        self.assert_result_queue_is_empty()
        self.assertEqual(
            result,
            ProgressResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                bytes_transferred=ref_bytes_transferred,
                total_transfer_size=self.size,
                timestamp=mock.ANY,
            ),
        )

    def test_on_done_success(self):
        subscriber = self.get_result_subscriber(DoneResultSubscriber)
        subscriber.on_done(self.future)
        result = self.get_queued_result()
        self.assert_result_queue_is_empty()
        self.assertEqual(
            result,
            SuccessResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
            ),
        )

    def test_on_done_failure(self):
        subscriber = self.get_result_subscriber(DoneResultSubscriber)
        subscriber.on_done(self.failure_future)
        result = self.get_queued_result()
        self.assert_result_queue_is_empty()
        self.assertEqual(
            result,
            FailureResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                exception=self.ref_exception,
            ),
        )

    def test_on_done_precondition_failed(self):
        subscriber = self.get_result_subscriber(DoneResultSubscriber)
        precondition_failed = HTTPClientError(
            request=mock.Mock(),
            response={
                'Error': {
                    'Code': 'PreconditionFailed'
                }
            },
            error='PreconditionFailed',
        )
        precondition_failed_future = self.get_failed_transfer_future(precondition_failed)
        subscriber.on_done(precondition_failed_future)
        result = self.get_queued_result()
        self.assert_result_queue_is_empty()
        self.assertEqual(
            result,
            SkipFileResult(transfer_type=mock.ANY, src=mock.ANY, dest=mock.ANY)
        )


    def test_on_done_unexpected_cancelled(self):
        subscriber = self.get_result_subscriber(DoneResultSubscriber)
        cancelled_exception = FatalError('some error')
        cancelled_future = self.get_failed_transfer_future(cancelled_exception)
        subscriber.on_done(cancelled_future)
        result = self.get_queued_result()
        self.assert_result_queue_is_empty()
        self.assertEqual(result, ErrorResult(exception=cancelled_exception))

    def test_on_done_cancelled_for_ctrl_c(self):
        subscriber = self.get_result_subscriber(DoneResultSubscriber)
        cancelled_exception = CancelledError('KeyboardInterrupt()')
        cancelled_future = self.get_failed_transfer_future(cancelled_exception)
        subscriber.on_done(cancelled_future)
        result = self.get_queued_result()
        self.assert_result_queue_is_empty()
        self.assertEqual(result, CtrlCResult(exception=cancelled_exception))


class ResultRecorderTest(unittest.TestCase):
    def setUp(self):
        self.transfer_type = 'upload'
        self.src = 'file'
        self.dest = 's3://mybucket/mykey'
        self.total_transfer_size = 20 * (1024**1024)  # 20MB
        self.warning_message = 'a dummy warning message'
        self.exception_message = 'a dummy exception message'
        self.exception = Exception(self.exception_message)
        self.result_recorder = ResultRecorder()
        self.result_recorder.start_time = 0

    def test_queued_result(self):
        self.result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                total_transfer_size=self.total_transfer_size,
            )
        )
        self.assertEqual(
            self.result_recorder.expected_bytes_transferred,
            self.total_transfer_size,
        )
        self.assertEqual(self.result_recorder.expected_files_transferred, 1)

    def test_multiple_queued_results(self):
        num_results = 5
        for i in range(num_results):
            self.result_recorder(
                QueuedResult(
                    transfer_type=self.transfer_type,
                    src=self.src + str(i),
                    dest=self.dest + str(i),
                    total_transfer_size=self.total_transfer_size,
                )
            )

        self.assertEqual(
            self.result_recorder.expected_bytes_transferred,
            num_results * self.total_transfer_size,
        )
        self.assertEqual(
            self.result_recorder.expected_files_transferred, num_results
        )

    def test_queued_result_with_no_full_transfer_size(self):
        self.result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                total_transfer_size=None,
            )
        )
        # Since we do not know how many bytes are expected to be transferred
        # do not incremenent the count as we have no idea how much it may be.
        self.assertEqual(self.result_recorder.expected_bytes_transferred, 0)
        self.assertEqual(self.result_recorder.expected_files_transferred, 1)

    def test_progress_result(self):
        self.result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                total_transfer_size=self.total_transfer_size,
            )
        )

        bytes_transferred = 1024 * 1024  # 1MB
        self.result_recorder(
            ProgressResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                bytes_transferred=bytes_transferred,
                total_transfer_size=self.total_transfer_size,
                timestamp=0,
            )
        )

        self.assertEqual(
            self.result_recorder.bytes_transferred, bytes_transferred
        )

    def test_multiple_progress_results(self):
        self.result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                total_transfer_size=self.total_transfer_size,
            )
        )

        bytes_transferred = 1024 * 1024  # 1MB
        num_results = 5
        for i in range(num_results):
            self.result_recorder(
                ProgressResult(
                    transfer_type=self.transfer_type,
                    src=self.src,
                    dest=self.dest,
                    bytes_transferred=bytes_transferred,
                    total_transfer_size=self.total_transfer_size,
                    timestamp=i,
                )
            )

        self.assertEqual(
            self.result_recorder.bytes_transferred,
            num_results * bytes_transferred,
        )

    def test_progress_result_with_no_known_transfer_size(self):
        self.result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                total_transfer_size=None,
            )
        )

        bytes_transferred = 1024 * 1024
        self.result_recorder(
            ProgressResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                bytes_transferred=bytes_transferred,
                total_transfer_size=None,
                timestamp=0,
            )
        )
        # Because the transfer size is still not known, update the
        # expected bytes transferred with what was actually transferred.
        self.assertEqual(
            self.result_recorder.bytes_transferred, bytes_transferred
        )
        self.assertEqual(
            self.result_recorder.expected_bytes_transferred, bytes_transferred
        )

    def test_progress_result_with_transfer_size_provided_during_progress(self):
        self.result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                total_transfer_size=None,
            )
        )

        bytes_transferred = 1024 * 1024
        self.result_recorder(
            ProgressResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                bytes_transferred=bytes_transferred,
                total_transfer_size=self.total_transfer_size,
                timestamp=0,
            )
        )

        self.assertEqual(
            self.result_recorder.bytes_transferred, bytes_transferred
        )
        # With the total size provided in the progress result, it should
        # accurately be reflected in the expected bytes transferred.
        self.assertEqual(
            self.result_recorder.expected_bytes_transferred,
            self.total_transfer_size,
        )

    def test_captures_start_time_on_queued(self):
        result_recorder = ResultRecorder()
        self.assertIsNone(result_recorder.start_time)
        result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                total_transfer_size=self.total_transfer_size,
            )
        )
        self.assertIsInstance(result_recorder.start_time, float)

    def test_progress_calculates_transfer_speed(self):
        start_time = 0
        self.result_recorder.start_time = start_time
        self.total_transfer_size = 10
        self.result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                total_transfer_size=self.total_transfer_size,
            )
        )
        # At this point nothing should have been uploaded so transfer speed
        # is zero
        self.assertEqual(self.result_recorder.bytes_transfer_speed, 0)

        self.result_recorder(
            ProgressResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                bytes_transferred=1,
                total_transfer_size=self.total_transfer_size,
                timestamp=(start_time + 1),
            )
        )

        # One bytes has been transferred in one second
        self.assertEqual(self.result_recorder.bytes_transfer_speed, 1)

        self.result_recorder(
            ProgressResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                bytes_transferred=4,
                total_transfer_size=self.total_transfer_size,
                timestamp=(start_time + 2),
            )
        )

        # Five bytes have been transferred in two seconds
        self.assertEqual(self.result_recorder.bytes_transfer_speed, 2.5)

        self.result_recorder(
            ProgressResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                bytes_transferred=1,
                total_transfer_size=self.total_transfer_size,
                timestamp=(start_time + 3),
            )
        )

        # Six bytes have been transferred in three seconds
        self.assertEqual(self.result_recorder.bytes_transfer_speed, 2.0)

    def test_success_result(self):
        self.result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                total_transfer_size=self.total_transfer_size,
            )
        )

        self.result_recorder(
            SuccessResult(
                transfer_type=self.transfer_type, src=self.src, dest=self.dest
            )
        )
        self.assertEqual(self.result_recorder.files_transferred, 1)
        self.assertEqual(self.result_recorder.files_failed, 0)

    def test_multiple_success_results(self):
        num_results = 5
        for i in range(num_results):
            self.result_recorder(
                QueuedResult(
                    transfer_type=self.transfer_type,
                    src=self.src + str(i),
                    dest=self.dest + str(i),
                    total_transfer_size=self.total_transfer_size,
                )
            )

        for i in range(num_results):
            self.result_recorder(
                SuccessResult(
                    transfer_type=self.transfer_type,
                    src=self.src + str(i),
                    dest=self.dest + str(i),
                )
            )

        self.assertEqual(self.result_recorder.files_transferred, num_results)
        self.assertEqual(self.result_recorder.files_failed, 0)

    def test_failure_result(self):
        self.result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                total_transfer_size=self.total_transfer_size,
            )
        )

        self.result_recorder(
            FailureResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                exception=self.exception,
            )
        )

        self.assertEqual(self.result_recorder.files_transferred, 1)
        self.assertEqual(self.result_recorder.files_failed, 1)
        self.assertEqual(
            self.result_recorder.bytes_failed_to_transfer,
            self.total_transfer_size,
        )
        self.assertEqual(self.result_recorder.bytes_transferred, 0)

    def test_multiple_failure_results(self):
        num_results = 5
        for i in range(num_results):
            self.result_recorder(
                QueuedResult(
                    transfer_type=self.transfer_type,
                    src=self.src + str(i),
                    dest=self.dest + str(i),
                    total_transfer_size=self.total_transfer_size,
                )
            )

        for i in range(num_results):
            self.result_recorder(
                FailureResult(
                    transfer_type=self.transfer_type,
                    src=self.src + str(i),
                    dest=self.dest + str(i),
                    exception=self.exception,
                )
            )

        self.assertEqual(self.result_recorder.files_transferred, num_results)
        self.assertEqual(self.result_recorder.files_failed, num_results)
        self.assertEqual(
            self.result_recorder.bytes_failed_to_transfer,
            self.total_transfer_size * num_results,
        )
        self.assertEqual(self.result_recorder.bytes_transferred, 0)

    def test_failure_result_mid_progress(self):
        self.result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                total_transfer_size=self.total_transfer_size,
            )
        )

        bytes_transferred = 1024 * 1024  # 1MB
        self.result_recorder(
            ProgressResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                bytes_transferred=bytes_transferred,
                total_transfer_size=self.total_transfer_size,
                timestamp=0,
            )
        )

        self.result_recorder(
            FailureResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                exception=self.exception,
            )
        )

        self.assertEqual(self.result_recorder.files_transferred, 1)
        self.assertEqual(self.result_recorder.files_failed, 1)
        self.assertEqual(
            self.result_recorder.bytes_failed_to_transfer,
            self.total_transfer_size - bytes_transferred,
        )
        self.assertEqual(
            self.result_recorder.bytes_transferred, bytes_transferred
        )

    def test_failure_result_still_did_not_know_transfer_size(self):
        self.result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                total_transfer_size=None,
            )
        )
        self.result_recorder(
            FailureResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                exception=self.exception,
            )
        )
        self.assertEqual(self.result_recorder.files_transferred, 1)
        self.assertEqual(self.result_recorder.files_failed, 1)
        # Because we never knew how many bytes to expect, do not make
        # any adjustments to bytes failed to transfer because it is impossible
        # to know that.
        self.assertEqual(self.result_recorder.bytes_failed_to_transfer, 0)

    def test_failure_result_and_learned_of_transfer_size_in_progress(self):
        self.result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                total_transfer_size=None,
            )
        )

        bytes_transferred = 1024 * 1024
        self.result_recorder(
            ProgressResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                bytes_transferred=bytes_transferred,
                total_transfer_size=self.total_transfer_size,
                timestamp=0,
            )
        )
        self.result_recorder(
            FailureResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                exception=self.exception,
            )
        )
        self.assertEqual(self.result_recorder.files_transferred, 1)
        self.assertEqual(self.result_recorder.files_failed, 1)
        # Since we knew how many bytes to expect at some point, it should
        # be accurately reflected in the amount failed to send when the
        # failure result is processed.
        self.assertEqual(
            self.result_recorder.bytes_failed_to_transfer,
            self.total_transfer_size - bytes_transferred,
        )

    def test_can_handle_results_with_no_dest(self):
        # This is just a quick smoke test to make sure that a result with
        # no destination like deletes can be handled for the lifecycle of
        # a transfer (i.e. being queued and finishing)
        self.result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=None,
                total_transfer_size=None,
            )
        )
        self.result_recorder(
            SuccessResult(
                transfer_type=self.transfer_type, src=self.src, dest=None
            )
        )
        self.assertEqual(self.result_recorder.expected_files_transferred, 1)
        self.assertEqual(self.result_recorder.files_transferred, 1)

    def test_warning_result(self):
        self.result_recorder(WarningResult(message=self.warning_message))
        self.assertEqual(self.result_recorder.files_warned, 1)

    def test_multiple_warning_results(self):
        num_results = 5
        for _ in range(num_results):
            self.result_recorder(WarningResult(message=self.warning_message))
        self.assertEqual(self.result_recorder.files_warned, num_results)

    def test_error_result(self):
        self.result_recorder(ErrorResult(exception=self.exception))
        self.assertEqual(self.result_recorder.errors, 1)

    def test_ctrl_c_result(self):
        self.result_recorder(CtrlCResult(Exception()))
        self.assertEqual(self.result_recorder.errors, 1)

    def test_expected_totals_are_final(self):
        self.result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                total_transfer_size=self.total_transfer_size,
            )
        )
        self.result_recorder(FinalTotalSubmissionsResult(1))
        self.assertEqual(
            self.result_recorder.final_expected_files_transferred, 1
        )
        self.assertTrue(self.result_recorder.expected_totals_are_final())

    def test_expected_totals_are_final_reaches_final_after_notification(self):
        self.result_recorder(FinalTotalSubmissionsResult(1))
        self.assertFalse(self.result_recorder.expected_totals_are_final())
        self.result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                total_transfer_size=self.total_transfer_size,
            )
        )
        self.assertTrue(self.result_recorder.expected_totals_are_final())

    def test_expected_totals_are_final_is_false_with_no_notification(self):
        self.assertIsNone(
            self.result_recorder.final_expected_files_transferred
        )
        self.assertFalse(self.result_recorder.expected_totals_are_final())
        self.result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=self.src,
                dest=self.dest,
                total_transfer_size=self.total_transfer_size,
            )
        )
        # It should still be None because it has not yet been notified
        # of finals.
        self.assertIsNone(
            self.result_recorder.final_expected_files_transferred
        )
        # This should remain False as well.
        self.assertFalse(self.result_recorder.expected_totals_are_final())

    def test_unknown_result_object(self):
        self.result_recorder(object())
        # Nothing should have been affected
        self.assertEqual(self.result_recorder.bytes_transferred, 0)
        self.assertEqual(self.result_recorder.expected_bytes_transferred, 0)
        self.assertEqual(self.result_recorder.expected_files_transferred, 0)
        self.assertEqual(self.result_recorder.files_transferred, 0)

    def test_result_with_unicode(self):
        unicode_source = '\u2713'
        self.result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=unicode_source,
                dest=self.dest,
                total_transfer_size=self.total_transfer_size,
            )
        )
        self.assertEqual(
            self.result_recorder.expected_bytes_transferred,
            self.total_transfer_size,
        )
        self.assertEqual(self.result_recorder.expected_files_transferred, 1)

    def test_result_with_encoded_unicode(self):
        unicode_source = '\u2713'.encode()
        self.result_recorder(
            QueuedResult(
                transfer_type=self.transfer_type,
                src=unicode_source,
                dest=self.dest,
                total_transfer_size=self.total_transfer_size,
            )
        )
        self.assertEqual(
            self.result_recorder.expected_bytes_transferred,
            self.total_transfer_size,
        )
        self.assertEqual(self.result_recorder.expected_files_transferred, 1)


class BaseResultPrinterTest(unittest.TestCase):
    def setUp(self):
        self.result_recorder = ResultRecorder()
        self.out_file = StringIO()
        self.error_file = StringIO()
        self.result_printer = ResultPrinter(
            result_recorder=self.result_recorder,
            out_file=self.out_file,
            error_file=self.error_file,
        )

    def get_progress_result(self):
        # NOTE: The actual values are not important for the purpose
        # of printing as the ResultPrinter only looks at the type and
        # the ResultRecorder to determine what to print out on progress.
        return ProgressResult(
            transfer_type=None,
            src=None,
            dest=None,
            bytes_transferred=None,
            total_transfer_size=None,
            timestamp=0,
        )


class TestResultPrinterMultiline(BaseResultPrinterTest):
    def setUp(self):
        # Overload the ResultPrinter to set oneline=False
        super().setUp()
        self.result_printer = ResultPrinter(
            result_recorder=self.result_recorder,
            out_file=self.out_file,
            error_file=self.error_file,
            oneline=False,
        )
        self.maxDiff = None

    def test_progress(self):
        mb = 1024 * 1024

        self.result_recorder.expected_bytes_transferred = 20 * mb
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.final_expected_files_transferred = 4
        self.result_recorder.bytes_transferred = mb
        self.result_recorder.files_transferred = 1

        progress_result = self.get_progress_result()
        self.result_printer(progress_result)
        ref_progress_statement = (
            'Completed 1.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) '
            'remaining\n'
        )
        self.assertEqual(self.out_file.getvalue(), ref_progress_statement)

    def test_success_with_progress(self):
        mb = 1024 * 1024

        progress_result = self.get_progress_result()

        # Add the first progress update and print it out
        self.result_recorder.expected_bytes_transferred = 20 * mb
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.final_expected_files_transferred = 4
        self.result_recorder.bytes_transferred = mb
        self.result_recorder.files_transferred = 1
        self.result_printer(progress_result)

        # Add a success result and print it out.
        transfer_type = 'upload'
        src = 'file'
        dest = 's3://mybucket/mykey'
        success_result = SuccessResult(
            transfer_type=transfer_type, src=src, dest=dest
        )

        self.result_recorder.files_transferred += 1
        self.result_printer(success_result)

        # The statement should consist of:
        # * The first progress statement
        # * The success statement
        # * And the progress again since the transfer is still ongoing
        # These lines will not have any padding
        ref_statement = (
            'Completed 1.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) remaining\n'
            'upload: file to s3://mybucket/mykey\n'
            'Completed 1.0 MiB/20.0 MiB (0 Bytes/s) with 2 file(s) remaining\n'
        )
        self.assertEqual(self.out_file.getvalue(), ref_statement)

    def test_progress_then_more_progress(self):
        mb = 1024 * 1024

        progress_result = self.get_progress_result()

        # Add the first progress update and print it out
        self.result_recorder.expected_bytes_transferred = 20 * mb
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.final_expected_files_transferred = 4
        self.result_recorder.bytes_transferred = mb
        self.result_recorder.files_transferred = 1

        self.result_printer(progress_result)
        ref_progress_statement = (
            'Completed 1.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) remaining\n'
        )
        self.assertEqual(self.out_file.getvalue(), ref_progress_statement)

        # Add the second progress update
        self.result_recorder.bytes_transferred += mb
        self.result_printer(progress_result)

        # The result should be the combination of the two
        ref_progress_statement = (
            'Completed 1.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) remaining\n'
            'Completed 2.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) remaining\n'
        )
        self.assertEqual(self.out_file.getvalue(), ref_progress_statement)


class TestResultPrinterFrequency(BaseResultPrinterTest):
    def setUp(self):
        # Overload the ResultPrinter to set oneline=False
        super().setUp()
        self.result_printer = ResultPrinter(
            result_recorder=self.result_recorder,
            out_file=self.out_file,
            error_file=self.error_file,
            oneline=False,
            frequency=2,
        )
        self.maxDiff = None

    def test_progress(self):
        mb = 1024 * 1024

        self.result_recorder.expected_bytes_transferred = 20 * mb
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.final_expected_files_transferred = 4
        self.result_recorder.bytes_transferred = mb
        self.result_recorder.files_transferred = 1

        progress_result = self.get_progress_result()
        self.result_printer(progress_result)
        ref_progress_statement = (
            'Completed 1.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) remaining\n'
        )
        self.assertEqual(self.out_file.getvalue(), ref_progress_statement)

    def test_progress_then_more_progress(self):
        now = time.time()

        mb = 1024 * 1024

        progress_result = self.get_progress_result()

        # Add the first progress update and print it out
        self.result_recorder.expected_bytes_transferred = 20 * mb
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.final_expected_files_transferred = 4
        self.result_recorder.bytes_transferred = mb
        self.result_recorder.files_transferred = 1

        self.result_printer(progress_result)
        ref_progress_statement = (
            'Completed 1.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) remaining\n'
        )
        self.assertEqual(self.out_file.getvalue(), ref_progress_statement)

        # Add the second progress update
        self.result_recorder.bytes_transferred += mb
        self.assertFalse(self.result_printer._first)
        with mock.patch(
            "awscli.customizations.s3.results.time.time"
        ) as mock_time:
            mock_time.return_value = now + 3
            self.result_printer(progress_result)
            # The result should be the combination of the two
            ref_progress_statement = (
                'Completed 1.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) remaining\n'
                'Completed 2.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) remaining\n'
            )
            self.assertEqual(self.out_file.getvalue(), ref_progress_statement)

    def test_progress_then_no_more_progress(self):
        """Check for progress at a time before the frequency minimum"""
        now = time.time()

        mb = 1024 * 1024

        progress_result = self.get_progress_result()

        # Add the first progress update and print it out
        self.result_recorder.expected_bytes_transferred = 20 * mb
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.final_expected_files_transferred = 4
        self.result_recorder.bytes_transferred = mb
        self.result_recorder.files_transferred = 1

        self.result_printer(progress_result)
        ref_progress_statement = (
            'Completed 1.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) remaining\n'
        )
        self.assertEqual(self.out_file.getvalue(), ref_progress_statement)

        # Add the second progress update
        self.result_recorder.bytes_transferred += mb
        self.assertFalse(self.result_printer._first)
        with mock.patch(
            "awscli.customizations.s3.results.time.time"
        ) as mock_time:
            mock_time.return_value = now + 1
            self.result_printer(progress_result)
            # The result should be the combination of the two
            ref_progress_statement = 'Completed 1.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) remaining\n'
            self.assertEqual(self.out_file.getvalue(), ref_progress_statement)


class TestResultPrinter(BaseResultPrinterTest):
    def test_unknown_result_object(self):
        self.result_printer(object())
        # Nothing should have been printed because of it.
        self.assertEqual(self.out_file.getvalue(), '')
        self.assertEqual(self.error_file.getvalue(), '')

    def test_progress(self):
        mb = 1024 * 1024

        self.result_recorder.expected_bytes_transferred = 20 * mb
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.final_expected_files_transferred = 4
        self.result_recorder.bytes_transferred = mb
        self.result_recorder.files_transferred = 1

        progress_result = self.get_progress_result()
        self.result_printer(progress_result)
        ref_progress_statement = (
            'Completed 1.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) '
            'remaining\r'
        )
        self.assertEqual(self.out_file.getvalue(), ref_progress_statement)

    def test_progress_with_no_expected_transfer_bytes(self):
        self.result_recorder.files_transferred = 1
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.final_expected_files_transferred = 4
        self.result_recorder.bytes_transferred = 0
        self.result_recorder.expected_bytes_transferred = 0

        progress_result = self.get_progress_result()
        self.result_printer(progress_result)
        ref_progress_statement = (
            'Completed 1 file(s) with 3 file(s) remaining\r'
        )
        self.assertEqual(self.out_file.getvalue(), ref_progress_statement)

    def test_progress_then_more_progress(self):
        mb = 1024 * 1024

        progress_result = self.get_progress_result()

        # Add the first progress update and print it out
        self.result_recorder.expected_bytes_transferred = 20 * mb
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.final_expected_files_transferred = 4
        self.result_recorder.bytes_transferred = mb
        self.result_recorder.files_transferred = 1

        self.result_printer(progress_result)
        ref_progress_statement = (
            'Completed 1.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) remaining\r'
        )
        self.assertEqual(self.out_file.getvalue(), ref_progress_statement)

        # Add the second progress update
        self.result_recorder.bytes_transferred += mb
        self.result_printer(progress_result)

        # The result should be the combination of the two
        ref_progress_statement = (
            'Completed 1.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) remaining\r'
            'Completed 2.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) remaining\r'
        )
        self.assertEqual(self.out_file.getvalue(), ref_progress_statement)

    def test_progress_still_calculating_totals(self):
        mb = 1024 * 1024

        self.result_recorder.expected_bytes_transferred = 20 * mb
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.bytes_transferred = mb
        self.result_recorder.files_transferred = 1

        progress_result = self.get_progress_result()
        self.result_printer(progress_result)
        ref_progress_statement = (
            'Completed 1.0 MiB/~20.0 MiB (0 Bytes/s) with ~3 file(s) '
            'remaining (calculating...)\r'
        )
        self.assertEqual(self.out_file.getvalue(), ref_progress_statement)

    def test_progress_still_calculating_totals_no_bytes(self):
        self.result_recorder.expected_bytes_transferred = 0
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.bytes_transferred = 0
        self.result_recorder.files_transferred = 1

        progress_result = self.get_progress_result()
        self.result_printer(progress_result)
        ref_progress_statement = (
            'Completed 1 file(s) with ~3 file(s) remaining (calculating...)\r'
        )
        self.assertEqual(self.out_file.getvalue(), ref_progress_statement)

    def test_progress_with_transfer_speed_reporting(self):
        mb = 1024 * 1024

        self.result_recorder.expected_bytes_transferred = 20 * mb
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.final_expected_files_transferred = 4
        self.result_recorder.bytes_transferred = mb
        self.result_recorder.files_transferred = 1
        self.result_recorder.bytes_transfer_speed = 1024 * 7

        progress_result = self.get_progress_result()
        self.result_printer(progress_result)
        ref_progress_statement = (
            'Completed 1.0 MiB/20.0 MiB (7.0 KiB/s) with 3 file(s) '
            'remaining\r'
        )
        self.assertEqual(self.out_file.getvalue(), ref_progress_statement)

    def test_success(self):
        transfer_type = 'upload'
        src = 'file'
        dest = 's3://mybucket/mykey'

        # Pretend that this is the final result in the result queue that
        # is processed.
        self.result_recorder.final_expected_files_transferred = 1
        self.result_recorder.expected_files_transferred = 1
        self.result_recorder.files_transferred = 1

        success_result = SuccessResult(
            transfer_type=transfer_type, src=src, dest=dest
        )

        self.result_printer(success_result)

        ref_success_statement = 'upload: file to s3://mybucket/mykey\n'
        self.assertEqual(self.out_file.getvalue(), ref_success_statement)

    def test_success_with_progress(self):
        mb = 1024 * 1024

        progress_result = self.get_progress_result()

        # Add the first progress update and print it out
        self.result_recorder.expected_bytes_transferred = 20 * mb
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.final_expected_files_transferred = 4
        self.result_recorder.bytes_transferred = mb
        self.result_recorder.files_transferred = 1
        self.result_printer(progress_result)

        # Add a success result and print it out.
        transfer_type = 'upload'
        src = 'file'
        dest = 's3://mybucket/mykey'
        success_result = SuccessResult(
            transfer_type=transfer_type, src=src, dest=dest
        )

        self.result_recorder.files_transferred += 1
        self.result_printer(success_result)

        # The statement should consist of:
        # * The first progress statement
        # * The success statement
        # * And the progress again since the transfer is still ongoing
        ref_statement = (
            'Completed 1.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) remaining\r'
            'upload: file to s3://mybucket/mykey                            \n'
            'Completed 1.0 MiB/20.0 MiB (0 Bytes/s) with 2 file(s) remaining\r'
        )
        self.assertEqual(self.out_file.getvalue(), ref_statement)

    def test_success_with_files_remaining(self):
        transfer_type = 'upload'
        src = 'file'
        dest = 's3://mybucket/mykey'

        mb = 1024 * 1024
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.files_transferred = 1
        self.result_recorder.expected_bytes_transferred = 4 * mb
        self.result_recorder.bytes_transferred = mb

        success_result = SuccessResult(
            transfer_type=transfer_type, src=src, dest=dest
        )

        self.result_printer(success_result)

        ref_success_statement = (
            'upload: file to s3://mybucket/mykey\n'
            'Completed 1.0 MiB/~4.0 MiB (0 Bytes/s) with ~3 file(s) '
            'remaining (calculating...)\r'
        )
        self.assertEqual(self.out_file.getvalue(), ref_success_statement)

    def test_success_but_no_expected_files_transferred_provided(self):
        transfer_type = 'upload'
        src = 'file'
        dest = 's3://mybucket/mykey'

        mb = 1024 * 1024
        self.result_recorder.expected_files_transferred = 1
        self.result_recorder.files_transferred = 1
        self.result_recorder.expected_bytes_transferred = mb
        self.result_recorder.bytes_transferred = mb

        success_result = SuccessResult(
            transfer_type=transfer_type, src=src, dest=dest
        )

        self.result_printer(success_result)

        ref_success_statement = (
            'upload: file to s3://mybucket/mykey\n'
            'Completed 1.0 MiB/~1.0 MiB (0 Bytes/s) with ~0 file(s) '
            'remaining (calculating...)\r'
        )
        self.assertEqual(self.out_file.getvalue(), ref_success_statement)

    def test_success_for_delete(self):
        transfer_type = 'delete'
        src = 's3://mybucket/mykey'

        # Pretend that this is the final result in the result queue that
        # is processed.
        self.result_recorder.final_expected_files_transferred = 1
        self.result_recorder.expected_files_transferred = 1
        self.result_recorder.files_transferred = 1

        success_result = SuccessResult(
            transfer_type=transfer_type, src=src, dest=None
        )

        self.result_printer(success_result)

        ref_success_statement = 'delete: s3://mybucket/mykey\n'
        self.assertEqual(self.out_file.getvalue(), ref_success_statement)

    def test_delete_success_with_files_remaining(self):
        transfer_type = 'delete'
        src = 's3://mybucket/mykey'

        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.files_transferred = 1

        success_result = SuccessResult(
            transfer_type=transfer_type, src=src, dest=None
        )

        self.result_printer(success_result)

        ref_success_statement = (
            'delete: s3://mybucket/mykey\n'
            'Completed 1 file(s) with ~3 file(s) remaining (calculating...)\r'
        )
        self.assertEqual(self.out_file.getvalue(), ref_success_statement)

    def test_delete_success_but_no_expected_files_transferred_provided(self):
        transfer_type = 'delete'
        src = 's3://mybucket/mykey'

        self.result_recorder.expected_files_transferred = 1
        self.result_recorder.files_transferred = 1

        success_result = SuccessResult(
            transfer_type=transfer_type, src=src, dest=None
        )

        self.result_printer(success_result)

        ref_success_statement = (
            'delete: s3://mybucket/mykey\n'
            'Completed 1 file(s) with ~0 file(s) remaining (calculating...)\r'
        )
        self.assertEqual(self.out_file.getvalue(), ref_success_statement)

    def test_failure(self):
        transfer_type = 'upload'
        src = 'file'
        dest = 's3://mybucket/mykey'

        # Pretend that this is the final result in the result queue that
        # is processed.
        self.result_recorder.final_expected_files_transferred = 1
        self.result_recorder.expected_files_transferred = 1
        self.result_recorder.files_transferred = 1

        failure_result = FailureResult(
            transfer_type=transfer_type,
            src=src,
            dest=dest,
            exception=Exception('my exception'),
        )

        self.result_printer(failure_result)

        ref_failure_statement = (
            'upload failed: file to s3://mybucket/mykey my exception\n'
        )
        self.assertEqual(self.error_file.getvalue(), ref_failure_statement)
        self.assertEqual(self.out_file.getvalue(), '')

    def test_failure_with_files_remaining(self):
        shared_file = self.out_file
        self.result_printer = ResultPrinter(
            result_recorder=self.result_recorder,
            out_file=shared_file,
            error_file=shared_file,
        )

        transfer_type = 'upload'
        src = 'file'
        dest = 's3://mybucket/mykey'

        mb = 1024 * 1024
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.files_transferred = 1
        self.result_recorder.expected_bytes_transferred = 4 * mb
        self.result_recorder.bytes_transferred = mb

        failure_result = FailureResult(
            transfer_type=transfer_type,
            src=src,
            dest=dest,
            exception=Exception('my exception'),
        )

        self.result_printer(failure_result)

        ref_statement = (
            'upload failed: file to s3://mybucket/mykey my exception\n'
            'Completed 1.0 MiB/~4.0 MiB (0 Bytes/s) with ~3 file(s) '
            'remaining (calculating...)\r'
        )
        self.assertEqual(self.out_file.getvalue(), ref_statement)

    def test_failure_but_no_expected_files_transferred_provided(self):
        shared_file = self.out_file
        self.result_printer = ResultPrinter(
            result_recorder=self.result_recorder,
            out_file=shared_file,
            error_file=shared_file,
        )

        transfer_type = 'upload'
        src = 'file'
        dest = 's3://mybucket/mykey'

        mb = 1024 * 1024
        self.result_recorder.expected_files_transferred = 1
        self.result_recorder.files_transferred = 1
        self.result_recorder.expected_bytes_transferred = mb
        self.result_recorder.bytes_transferred = mb

        failure_result = FailureResult(
            transfer_type=transfer_type,
            src=src,
            dest=dest,
            exception=Exception('my exception'),
        )

        self.result_printer(failure_result)

        ref_statement = (
            'upload failed: file to s3://mybucket/mykey my exception\n'
            'Completed 1.0 MiB/~1.0 MiB (0 Bytes/s) with ~0 file(s) '
            'remaining (calculating...)\r'
        )
        self.assertEqual(self.out_file.getvalue(), ref_statement)

    def test_failure_with_progress(self):
        # Make errors and regular outprint go to the same file to track order.
        shared_file = self.out_file
        self.result_printer = ResultPrinter(
            result_recorder=self.result_recorder,
            out_file=shared_file,
            error_file=shared_file,
        )

        mb = 1024 * 1024

        progress_result = self.get_progress_result()

        # Add the first progress update and print it out
        self.result_recorder.expected_bytes_transferred = 20 * mb
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.final_expected_files_transferred = 4
        self.result_recorder.bytes_transferred = mb
        self.result_recorder.files_transferred = 1
        self.result_printer(progress_result)

        # Add a success result and print it out.
        transfer_type = 'upload'
        src = 'file'
        dest = 's3://mybucket/mykey'
        failure_result = FailureResult(
            transfer_type=transfer_type,
            src=src,
            dest=dest,
            exception=Exception('my exception'),
        )

        self.result_recorder.bytes_failed_to_transfer = 3 * mb
        self.result_recorder.files_transferred += 1
        self.result_printer(failure_result)

        # The statement should consist of:
        # * The first progress statement
        # * The failure statement
        # * And the progress again since the transfer is still ongoing
        ref_statement = (
            'Completed 1.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) remaining\r'
            'upload failed: file to s3://mybucket/mykey my exception        \n'
            'Completed 4.0 MiB/20.0 MiB (0 Bytes/s) with 2 file(s) remaining\r'
        )
        self.assertEqual(shared_file.getvalue(), ref_statement)

    def test_failure_for_delete(self):
        transfer_type = 'delete'
        src = 's3://mybucket/mykey'

        # Pretend that this is the final result in the result queue that
        # is processed.
        self.result_recorder.final_expected_files_transferred = 1
        self.result_recorder.expected_files_transferred = 1
        self.result_recorder.files_transferred = 1

        failure_result = FailureResult(
            transfer_type=transfer_type,
            src=src,
            dest=None,
            exception=Exception('my exception'),
        )

        self.result_printer(failure_result)

        ref_failure_statement = (
            'delete failed: s3://mybucket/mykey my exception\n'
        )
        self.assertEqual(self.error_file.getvalue(), ref_failure_statement)
        self.assertEqual(self.out_file.getvalue(), '')

    def test_delete_failure_with_files_remaining(self):
        shared_file = self.out_file
        self.result_printer = ResultPrinter(
            result_recorder=self.result_recorder,
            out_file=shared_file,
            error_file=shared_file,
        )

        transfer_type = 'delete'
        src = 's3://mybucket/mykey'

        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.files_transferred = 1

        failure_result = FailureResult(
            transfer_type=transfer_type,
            src=src,
            dest=None,
            exception=Exception('my exception'),
        )

        self.result_printer(failure_result)

        ref_statement = (
            'delete failed: s3://mybucket/mykey my exception\n'
            'Completed 1 file(s) with ~3 file(s) remaining (calculating...)\r'
        )
        self.assertEqual(self.out_file.getvalue(), ref_statement)

    def test_delete_failure_but_no_expected_files_transferred_provided(self):
        shared_file = self.out_file
        self.result_printer = ResultPrinter(
            result_recorder=self.result_recorder,
            out_file=shared_file,
            error_file=shared_file,
        )

        transfer_type = 'delete'
        src = 's3://mybucket/mykey'

        self.result_recorder.expected_files_transferred = 1
        self.result_recorder.files_transferred = 1

        failure_result = FailureResult(
            transfer_type=transfer_type,
            src=src,
            dest=None,
            exception=Exception('my exception'),
        )

        self.result_printer(failure_result)

        ref_statement = (
            'delete failed: s3://mybucket/mykey my exception\n'
            'Completed 1 file(s) with ~0 file(s) remaining (calculating...)\r'
        )
        self.assertEqual(self.out_file.getvalue(), ref_statement)

    def test_warning(self):
        # Pretend that this is the final result in the result queue that
        # is processed.
        self.result_recorder.final_expected_files_transferred = 1
        self.result_recorder.expected_files_transferred = 1
        self.result_recorder.files_transferred = 1

        self.result_printer(WarningResult('warning: my warning'))
        ref_warning_statement = 'warning: my warning\n'
        self.assertEqual(self.error_file.getvalue(), ref_warning_statement)
        self.assertEqual(self.out_file.getvalue(), '')

    def test_warning_with_progress(self):
        # Make errors and regular outprint go to the same file to track order.
        shared_file = self.out_file
        self.result_printer = ResultPrinter(
            result_recorder=self.result_recorder,
            out_file=shared_file,
            error_file=shared_file,
        )

        mb = 1024 * 1024

        progress_result = self.get_progress_result()

        # Add the first progress update and print it out
        self.result_recorder.expected_bytes_transferred = 20 * mb
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.final_expected_files_transferred = 4
        self.result_recorder.bytes_transferred = mb
        self.result_recorder.files_transferred = 1
        self.result_printer(progress_result)

        self.result_printer(WarningResult('warning: my warning'))

        # The statement should consist of:
        # * The first progress statement
        # * The warning statement
        # * And the progress again since the transfer is still ongoing
        ref_statement = (
            'Completed 1.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) remaining\r'
            'warning: my warning                                            \n'
            'Completed 1.0 MiB/20.0 MiB (0 Bytes/s) with 3 file(s) remaining\r'
        )

        self.assertEqual(shared_file.getvalue(), ref_statement)

    def test_error(self):
        self.result_printer(ErrorResult(Exception('my exception')))
        ref_error_statement = 'fatal error: my exception\n'
        self.assertEqual(self.error_file.getvalue(), ref_error_statement)

    def test_ctrl_c_error(self):
        self.result_printer(CtrlCResult(Exception()))
        ref_error_statement = 'cancelled: ctrl-c received\n'
        self.assertEqual(self.error_file.getvalue(), ref_error_statement)

    def test_error_while_progress(self):
        mb = 1024 * 1024
        self.result_recorder.expected_bytes_transferred = 20 * mb
        self.result_recorder.expected_files_transferred = 4
        self.result_recorder.final_expected_files_transferred = 4
        self.result_recorder.bytes_transferred = mb
        self.result_recorder.files_transferred = 1

        self.result_printer(ErrorResult(Exception('my exception')))
        ref_error_statement = 'fatal error: my exception\n'
        # Even though there was progress, we do not want to print the
        # progress because errors are really only seen when the entire
        # s3 command fails.
        self.assertEqual(self.error_file.getvalue(), ref_error_statement)

    def test_dry_run(self):
        result = DryRunResult(
            transfer_type='upload',
            src='s3://mybucket/key',
            dest='./local/file',
        )
        self.result_printer(result)
        expected = '(dryrun) upload: s3://mybucket/key to ./local/file\n'
        self.assertEqual(self.out_file.getvalue(), expected)

    def test_final_total_notification_with_no_more_expected_progress(self):
        transfer_type = 'upload'
        src = 'file'
        dest = 's3://mybucket/mykey'

        mb = 1024 * 1024
        self.result_recorder.expected_files_transferred = 1
        self.result_recorder.files_transferred = 1
        self.result_recorder.expected_bytes_transferred = mb
        self.result_recorder.bytes_transferred = mb

        success_result = SuccessResult(
            transfer_type=transfer_type, src=src, dest=dest
        )

        self.result_printer(success_result)

        ref_success_statement = (
            'upload: file to s3://mybucket/mykey\n'
            'Completed 1.0 MiB/~1.0 MiB (0 Bytes/s) with ~0 file(s) '
            'remaining (calculating...)\r'
        )
        self.assertEqual(self.out_file.getvalue(), ref_success_statement)

        # Now the result recorder/printer is notified it was just
        # there will be no more queueing. Therefore it should
        # clear out remaining progress if the expected number of files
        # transferred is equal to the number of files that has completed
        # because this is the final task meaning we want to clear any progress
        # that is displayed.
        self.result_recorder.final_expected_files_transferred = 1
        self.result_printer(FinalTotalSubmissionsResult(1))
        ref_success_statement = (
            'upload: file to s3://mybucket/mykey\n'
            'Completed 1.0 MiB/~1.0 MiB (0 Bytes/s) '
            'with ~0 file(s) remaining (calculating...)\r'
            '                                             '
            '                                    \n'
        )
        self.assertEqual(self.out_file.getvalue(), ref_success_statement)

    def test_final_total_does_not_print_out_newline_for_no_transfers(self):
        self.result_recorder.final_expected_files_transferred = 0
        self.result_printer(FinalTotalSubmissionsResult(0))
        self.assertEqual(self.out_file.getvalue(), '')

    def test_print_unicode_success_src_and_dest(self):
        # Pretend that this is the final result in the result queue that
        # is processed.
        self.result_recorder.final_expected_files_transferred = 1
        self.result_recorder.expected_files_transferred = 1
        self.result_recorder.files_transferred = 1

        result = SuccessResult(
            transfer_type='upload',
            src='/tmp/\u2713',
            dest='s3://mybucket/mykey',
        )
        self.result_printer(result)
        expected = 'upload: /tmp/\u2713 to s3://mybucket/mykey\n'
        self.assertEqual(self.out_file.getvalue(), expected)

    def test_print_unicode_success_src(self):
        # Pretend that this is the final result in the result queue that
        # is processed.
        self.result_recorder.final_expected_files_transferred = 1
        self.result_recorder.expected_files_transferred = 1
        self.result_recorder.files_transferred = 1

        result = SuccessResult(
            transfer_type='delete', src='s3://mybucket/tmp/\u2713', dest=None
        )
        self.result_printer(result)
        expected = 'delete: s3://mybucket/tmp/\u2713\n'
        self.assertEqual(self.out_file.getvalue(), expected)

    def test_print_unicode_dryrun(self):
        result = DryRunResult(
            transfer_type='upload',
            src='s3://mybucket/\u2713',
            dest='./local/file',
        )
        self.result_printer(result)
        expected = '(dryrun) upload: s3://mybucket/\u2713 to ./local/file\n'
        self.assertEqual(self.out_file.getvalue(), expected)

    def test_print_unicode_failure(self):
        transfer_type = 'upload'
        src = '\u2713'
        dest = 's3://mybucket/mykey'

        # Pretend that this is the final result in the result queue that
        # is processed.
        self.result_recorder.final_expected_files_transferred = 1
        self.result_recorder.expected_files_transferred = 1
        self.result_recorder.files_transferred = 1

        failure_result = FailureResult(
            transfer_type=transfer_type,
            src=src,
            dest=dest,
            exception=Exception('my exception'),
        )

        self.result_printer(failure_result)

        ref_failure_statement = (
            'upload failed: \u2713 to s3://mybucket/mykey my exception\n'
        )
        self.assertEqual(self.error_file.getvalue(), ref_failure_statement)
        self.assertEqual(self.out_file.getvalue(), '')

    def test_print_unicode_warning(self):
        # Pretend that this is the final result in the result queue that
        # is processed.
        self.result_recorder.final_expected_files_transferred = 1
        self.result_recorder.expected_files_transferred = 1
        self.result_recorder.files_transferred = 1

        self.result_printer(WarningResult('warning: unicode exists \u2713'))
        ref_warning_statement = 'warning: unicode exists \u2713\n'
        self.assertEqual(self.error_file.getvalue(), ref_warning_statement)
        self.assertEqual(self.out_file.getvalue(), '')

    def test_print_unicode_error(self):
        self.result_printer(ErrorResult(Exception('unicode exists \u2713')))
        ref_error_statement = 'fatal error: unicode exists \u2713\n'
        self.assertEqual(self.error_file.getvalue(), ref_error_statement)

    def test_does_not_print_skipped_file(self):
        transfer_type = 'upload'
        src = 'file'
        dest = 's3://mybucket/mykey'
        skip_file_result = SkipFileResult(
            transfer_type=transfer_type, src=src, dest=dest
        )

        # Pretend that this is the final result in the result queue that
        # is processed.
        self.result_recorder.final_expected_files_transferred = 1
        self.result_recorder.expected_files_transferred = 1
        self.result_recorder.files_skipped = 1

        self.result_printer(skip_file_result)
        self.assertEqual(self.out_file.getvalue(), '')


class TestNoProgressResultPrinter(BaseResultPrinterTest):
    def setUp(self):
        super().setUp()
        self.result_printer = NoProgressResultPrinter(
            result_recorder=self.result_recorder,
            out_file=self.out_file,
            error_file=self.error_file,
        )

    def test_does_not_print_progress_result(self):
        progress_result = self.get_progress_result()
        self.result_printer(progress_result)
        self.assertEqual(self.out_file.getvalue(), '')

    def test_does_print_sucess_result(self):
        transfer_type = 'upload'
        src = 'file'
        dest = 's3://mybucket/mykey'
        success_result = SuccessResult(
            transfer_type=transfer_type, src=src, dest=dest
        )

        self.result_printer(success_result)
        expected_message = 'upload: file to s3://mybucket/mykey\n'
        self.assertEqual(self.out_file.getvalue(), expected_message)

    def test_print_failure_result(self):
        transfer_type = 'upload'
        src = 'file'
        dest = 's3://mybucket/mykey'
        failure_result = FailureResult(
            transfer_type=transfer_type,
            src=src,
            dest=dest,
            exception=Exception('my exception'),
        )

        self.result_printer(failure_result)

        ref_failure_statement = (
            'upload failed: file to s3://mybucket/mykey my exception\n'
        )
        self.assertEqual(self.error_file.getvalue(), ref_failure_statement)

    def test_print_warnings_result(self):
        self.result_printer(WarningResult('warning: my warning'))
        ref_warning_statement = 'warning: my warning\n'
        self.assertEqual(self.error_file.getvalue(), ref_warning_statement)

    def test_final_total_does_not_try_to_clear_empty_progress(self):
        transfer_type = 'upload'
        src = 'file'
        dest = 's3://mybucket/mykey'

        mb = 1024 * 1024
        self.result_recorder.expected_files_transferred = 1
        self.result_recorder.files_transferred = 1
        self.result_recorder.expected_bytes_transferred = mb
        self.result_recorder.bytes_transferred = mb

        success_result = SuccessResult(
            transfer_type=transfer_type, src=src, dest=dest
        )
        self.result_printer(success_result)
        ref_statement = 'upload: file to s3://mybucket/mykey\n'
        self.assertEqual(self.out_file.getvalue(), ref_statement)

        self.result_recorder.final_expected_files_transferred = 1
        self.result_printer(FinalTotalSubmissionsResult(1))
        self.assertEqual(self.out_file.getvalue(), ref_statement)


class TestOnlyShowErrorsResultPrinter(BaseResultPrinterTest):
    def setUp(self):
        super().setUp()
        self.result_printer = OnlyShowErrorsResultPrinter(
            result_recorder=self.result_recorder,
            out_file=self.out_file,
            error_file=self.error_file,
        )

    def test_does_not_print_progress_result(self):
        progress_result = self.get_progress_result()
        self.result_printer(progress_result)
        self.assertEqual(self.out_file.getvalue(), '')

    def test_does_not_print_sucess_result(self):
        transfer_type = 'upload'
        src = 'file'
        dest = 's3://mybucket/mykey'
        success_result = SuccessResult(
            transfer_type=transfer_type, src=src, dest=dest
        )

        self.result_printer(success_result)
        self.assertEqual(self.out_file.getvalue(), '')

    def test_print_failure_result(self):
        transfer_type = 'upload'
        src = 'file'
        dest = 's3://mybucket/mykey'
        failure_result = FailureResult(
            transfer_type=transfer_type,
            src=src,
            dest=dest,
            exception=Exception('my exception'),
        )

        self.result_printer(failure_result)

        ref_failure_statement = (
            'upload failed: file to s3://mybucket/mykey my exception\n'
        )
        self.assertEqual(self.error_file.getvalue(), ref_failure_statement)

    def test_print_warnings_result(self):
        self.result_printer(WarningResult('warning: my warning'))
        ref_warning_statement = 'warning: my warning\n'
        self.assertEqual(self.error_file.getvalue(), ref_warning_statement)

    def test_final_total_does_not_try_to_clear_empty_progress(self):
        transfer_type = 'upload'
        src = 'file'
        dest = 's3://mybucket/mykey'

        mb = 1024 * 1024
        self.result_recorder.expected_files_transferred = 1
        self.result_recorder.files_transferred = 1
        self.result_recorder.expected_bytes_transferred = mb
        self.result_recorder.bytes_transferred = mb

        success_result = SuccessResult(
            transfer_type=transfer_type, src=src, dest=dest
        )
        self.result_printer(success_result)
        ref_statement = ''
        self.assertEqual(self.out_file.getvalue(), ref_statement)

        self.result_recorder.final_expected_files_transferred = 1
        self.result_printer(FinalTotalSubmissionsResult(1))
        # The final total submission result should be a noop and
        # not print anything out.
        self.assertEqual(self.out_file.getvalue(), ref_statement)


class TestResultProcessor(unittest.TestCase):
    def setUp(self):
        self.result_queue = queue.Queue()
        self.result_recorder = mock.Mock()
        self.result_printer = mock.Mock()
        self.results_handled = []

        self.result_processor = ResultProcessor(
            self.result_queue, [self.results_handled.append]
        )

    def _handle_result_with_exception(self, result):
        raise Exception()

    def test_run(self):
        transfer_type = 'upload'
        src = 'src'
        dest = 'dest'
        total_transfer_size = 1024 * 1024
        results_to_process = [
            QueuedResult(transfer_type, src, dest, total_transfer_size),
            SuccessResult(transfer_type, src, dest),
        ]
        results_with_shutdown = results_to_process + [ShutdownThreadRequest()]

        for result in results_with_shutdown:
            self.result_queue.put(result)
        self.result_processor.run()

        self.assertEqual(self.results_handled, results_to_process)

    def test_run_without_result_handlers(self):
        transfer_type = 'upload'
        src = 'src'
        dest = 'dest'
        total_transfer_size = 1024 * 1024
        results_to_process = [
            QueuedResult(transfer_type, src, dest, total_transfer_size),
            SuccessResult(transfer_type, src, dest),
        ]
        results_with_shutdown = results_to_process + [ShutdownThreadRequest()]

        for result in results_with_shutdown:
            self.result_queue.put(result)
        self.result_processor = ResultProcessor(self.result_queue)
        self.result_processor.run()

        # Ensure that the entire result queue got processed even though
        # there was no handlers provided.
        self.assertTrue(self.result_queue.empty())

    def test_exception_handled_in_loop(self):
        transfer_type = 'upload'
        src = 'src'
        dest = 'dest'
        total_transfer_size = 1024 * 1024
        results_to_process = [
            QueuedResult(transfer_type, src, dest, total_transfer_size),
            SuccessResult(transfer_type, src, dest),
        ]
        results_with_shutdown = results_to_process + [ShutdownThreadRequest()]

        for result in results_with_shutdown:
            self.result_queue.put(result)

        results_handled_after_exception = []
        self.result_processor = ResultProcessor(
            self.result_queue,
            [
                self.results_handled.append,
                self._handle_result_with_exception,
                results_handled_after_exception.append,
            ],
        )

        self.result_processor.run()

        self.assertEqual(self.results_handled, results_to_process)
        # The exception happens in the second handler, the exception being
        # thrown should result in the first handler and the ResultProcessor
        # continuing to process through the result queue despite the exception.
        # However, any handlers after the exception should not be run just
        # in case order of the handlers mattering and an unhandled exception
        # in one affects another handler.
        self.assertEqual(results_handled_after_exception, results_to_process)

    def test_does_not_handle_results_after_receiving_error_result(self):
        transfer_type = 'upload'
        src = 'src'
        dest = 'dest'
        results_to_be_handled = [
            SuccessResult(transfer_type, src, dest),
            ErrorResult(Exception('my exception')),
        ]
        result_not_to_be_handled = [
            ErrorResult(Exception('my second exception'))
        ]
        results_with_shutdown = (
            results_to_be_handled
            + result_not_to_be_handled
            + [ShutdownThreadRequest()]
        )

        for result in results_with_shutdown:
            self.result_queue.put(result)

        self.result_processor.run()
        # Only the results including and before the first the ErrorResult
        # should be handled. Any results after the first ErrorResult should
        # be ignored because ErrorResults are considered fatal meaning the
        # ResultProcessor needs to consume through the rest of result queue
        # to shutdown as quickly as possible.
        self.assertEqual(self.results_handled, results_to_be_handled)

    def test_does_not_process_results_after_shutdown(self):
        transfer_type = 'upload'
        src = 'src'
        dest = 'dest'
        total_transfer_size = 1024 * 1024
        results_to_process = [
            QueuedResult(transfer_type, src, dest, total_transfer_size),
            SuccessResult(transfer_type, src, dest),
        ]
        results_with_shutdown = results_to_process + [
            ShutdownThreadRequest(),
            WarningResult('my warning'),
        ]

        for result in results_with_shutdown:
            self.result_queue.put(result)
        self.result_processor.run()
        # Because a ShutdownThreadRequest was sent the processor should
        # not have processed anymore results stored after it.
        self.assertEqual(self.results_handled, results_to_process)


class TestCommandResultRecorder(unittest.TestCase):
    def setUp(self):
        self.result_queue = queue.Queue()
        self.result_recorder = ResultRecorder()
        self.result_processor = ResultProcessor(
            self.result_queue, [self.result_recorder]
        )
        self.command_result_recorder = CommandResultRecorder(
            self.result_queue, self.result_recorder, self.result_processor
        )

        self.transfer_type = 'upload'
        self.src = 'file'
        self.dest = 's3://mybucket/mykey'
        self.total_transfer_size = 20 * (1024**1024)

    def test_success(self):
        with self.command_result_recorder:
            self.result_queue.put(
                QueuedResult(
                    transfer_type=self.transfer_type,
                    src=self.src,
                    dest=self.dest,
                    total_transfer_size=self.total_transfer_size,
                )
            )
            self.result_queue.put(
                SuccessResult(
                    transfer_type=self.transfer_type,
                    src=self.src,
                    dest=self.dest,
                )
            )
        self.assertEqual(
            self.command_result_recorder.get_command_result(), (0, 0)
        )

    def test_fail(self):
        with self.command_result_recorder:
            self.result_queue.put(
                QueuedResult(
                    transfer_type=self.transfer_type,
                    src=self.src,
                    dest=self.dest,
                    total_transfer_size=self.total_transfer_size,
                )
            )
            self.result_queue.put(
                FailureResult(
                    transfer_type=self.transfer_type,
                    src=self.src,
                    dest=self.dest,
                    exception=Exception('my exception'),
                )
            )
        self.assertEqual(
            self.command_result_recorder.get_command_result(), (1, 0)
        )

    def test_warning(self):
        with self.command_result_recorder:
            self.result_queue.put(WarningResult(message='my warning'))
        self.assertEqual(
            self.command_result_recorder.get_command_result(), (0, 1)
        )

    def test_error(self):
        with self.command_result_recorder:
            raise Exception('my exception')
        self.assertEqual(
            self.command_result_recorder.get_command_result(), (1, 0)
        )

    def test_notify_total_submissions(self):
        total = 5
        self.command_result_recorder.notify_total_submissions(total)
        self.assertEqual(
            self.result_queue.get(), FinalTotalSubmissionsResult(total)
        )
